####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import os
import numpy as np
import typing as tp
import weakref as wref
import _paramonte as pm
import _pmutils as pmutils

Struct = pmutils.Struct
newline = pmutils.newline

####################################################################################################################################
#### BasePlot class
####################################################################################################################################

class BasePlot:
    """

    This is the class for generating instances of
    basic plots with minimally one (X)-axis. It serves as the
    superclass for a wide variety of other multi-axes ParaMonte plots.

        **Parameters**

            plotType

                A string indicating the name of the plot that is to be constructed.

            dataFrame (optional)

                A pandas dataFrame whose data will be plotted.

            methodName (optional)

                The name of the ParaMonte sample requesting the BasePlot.

            reportEnabled (optional)

                A boolean whose value indicates whether guidelines should be 
                printed in the standard output.

            resetPlot (optional)

                A function that resets the properties of the plot as desired 
                from outside. If provided, a pointer to this function will be
                saved for future internal usage.

        **Attributes**

            _dfref

                A weak reference to the input dataFrame whose data is used to 
                generate plots.

            rows

                A numeric vector that determines the rows of dataFrame
                to be visualized. It can be either:

                    1.  a ``range(start,stop,step)``, or,
                    2.  a list of row indices in ``dataFrame.index``.

                Examples:

                    1.  ``rows = range(17,7,-2)``
                    2.  ``rows = [i for i in range(7,17)]``

                If not provided, the default includes all rows of the dataframe.

            set

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``set()``
                        function of the seaborn library should be made or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set()`` function.

                        Example usage:

                            .. code-block:: python

                                set.kws.style = "darkgrid"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes (available only in 1D and 2D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``gca()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes3d (available only in 3D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``Axes3D()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes3d.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            figure

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``figure()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``figure()`` function of 
                        the matplotlib library.

                        Example usage:

                            .. code-block:: python

                                figure.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            legend (may not exist for some types of plots)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``legend()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``legend()`` function of 
                        the matplotlib library.

                        Example usage:

                        .. code-block:: python

                            legend.kws.labels = ["Variable1", "Variable2"]

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A legend will be added to a plot only if no color-mappings are
                requested in the plot.

            currentFig

                A structure whose attributes are the outputs of various plotting 
                tools used to make the current figure. These include the handle 
                to the current figure, the handle to the current axes in the plot, 
                the handle to the colorbar (if any exists), and other Python 
                plotting tools used to make to generate the figure.

        **Returns**

            An object of ``BasePlot`` class.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , plotType
                , dataFrame = None
                , methodName = "ParaMonte"
                , reportEnabled = True
                , resetPlot = None
                ):

            self._dfref = None if dataFrame is None else wref.ref(dataFrame)

            self._type = Struct()
            self._type.name = plotType

            plotTypeLower = plotType.lower()

            self._type.is3d         = "3"           in plotTypeLower
            self._type.isLine       = "line"        in plotTypeLower
            self._type.isScatter    = "scatter"     in plotTypeLower
            self._type.isHeatmap    = "heatmap"     in plotTypeLower
            self._type.isKdeplot1   = "kdeplot1"    in plotTypeLower
            self._type.isKdeplot2   = "kdeplot2"    in plotTypeLower
            self._type.isHistplot   = "histplot"    in plotTypeLower
            self._type.isJointplot  = "jointplot"   in plotTypeLower
            self._type.isEllipsoid  = "covmat"      in plotTypeLower
            self._type.isEllipsoid  = "cormat"      in plotTypeLower
            self._type.isGridPlot   = "grid"        in plotTypeLower
            self._type.isContour3   = "contour3"    in plotTypeLower
            self._type.isContourf   = "contourf"    in plotTypeLower
            self._type.isContour    = "contour"     in plotTypeLower and not (self._type.isContour3 or self._type.isContourf)

            self._type.is1d         = self._type.isKdeplot1 or self._type.isHistplot
            self._type.is2d         = not (self._type.isGridPlot or self._type.is1d or self._type.is3d)

            self._type.isDiffusionPlot = self._type.isContour or self._type.isContourf or self._type.isContour3

            self._methodName = methodName
            self._reportEnabled = reportEnabled

            self._progress = pm.utils.Progress  ( msg = "creating a " + self._type.name + " plot object from scratch... "
                                                , methodName = methodName
                                                , reportEnabled = reportEnabled
                                                , end = ""
                                                )

            self._reset()

            if resetPlot is None:
                self._resetPlot = self._reset
            else:
                self._resetPlot = resetPlot

    ################################################################################################################################
    #### reset
    ################################################################################################################################

    def reset   ( self
                , resetType = "soft"
                , **kwargs
                ): 
        """

        Reset the properties of the plot to the original default settings.
        Use this method when you change many attributes of the plot and
        you want to clean up and go back to the default settings.

            **Parameters**

                resetType (optional)

                    An optional string with possible value of ``"hard"``.
                    If provided, the plot object will be regenerated from scratch.
                    This includes reading the original data frame again and resetting
                    everything. If not provided, then only the plot settings will be
                    reset without reseting the dataFrame.

            **Returns**

                None

            **Example**

                .. code-block:: python
                    :linenos:

                    reset()         # reset the plot to the default settings
                    reset("soft")   # reset the plot to the default settings
                    reset("hard")   # regenerate the plot from scratch

        """

        try:
            self._resetPlot(resetType = resetType, plotNames = self._type.name) # call external reset function
        except:
            self._resetPlot(**kwargs) # call _reset(self)

    ################################################################################################################################
    #### _reset
    ################################################################################################################################

    def _reset(self):

        self.rows = None
        self._fnameOld = None # used in savefig() to generate unique random file names.

        self.set = Struct()
        self.set.enabled = True
        self.set.kws = Struct()
        if self._type.isKdeplot2 or self._type.isJointplot:
            self.set.kws.style = "ticks"
        elif self._type.is3d:
            self.set.kws.style = "white"
        else:
            self.set.kws.style = "darkgrid"

        self.figure = Struct()
        self.figure.enabled = True
        self.figure.kws = Struct()

        self.currentFig = Struct()

        if self._type.is3d:
            self.axes3d = Struct()
            self.axes3d.kws = Struct()
            self.axes3d.kws.alpha = 1
            self.axes3d.kws.visible = True
            #self.axes3d.enabled = self.figure.enabled
        else:
            self.axes = Struct()
            self.axes.kws = Struct()
            self.axes.kws.alpha = 1
            self.axes.kws.visible = True
            #self.axes.enabled = self.figure.enabled

        if not (self._type.isGridPlot or self._type.isHeatmap or self._type.isDiffusionPlot):
            self.legend = Struct()
            self.legend.enabled = False
            self.legend.kws = Struct()

    ################################################################################################################################
    #### savefig
    ################################################################################################################################

    def savefig ( self
                , reself = False
                , **savefig_kws
                ):
        """

        Export the current figure to external file.

            **Parameters**

                reself

                    A logical variable. If ``True``, the path to 
                    the output generated file will be returned.
                    The default value is ``False``.

                **savefig_kws (optional)

                The set of input arguments to the ``savefig()`` function 
                of the matplotlib library, including the ``fname``. If 
                the file name is not provided, a unique random filename 
                will be generated and used to save the figure to an 
                output file.

                Example:

                .. code-block:: python

                    savefig(fname = "gridplot.png")

        **Returns**

            None

        **Example**

            .. code-block:: python

                savefig() # use a unique random filename to output the plot.
                savefig(fname = "thisPlot.png")
                savefig(pad_inches = 0.0, bbox_inches = "tight")

        """

        if "fname" not in savefig_kws.keys() or (isinstance(self._fnameOld,str) and isinstance(savefig_kws["fname"],str) and savefig_kws["fname"]==self._fnameOld):
            savefig_kws["fname"] = os.path.join( os.getcwd(), pm.utils.getRandomFilePrefix(prefix=self._type.name+"_") + ".png" )
            self._fnameOld = savefig_kws["fname"]

        import matplotlib.pyplot as plt # xxx how is this figure activated?
        self._progress.note( msg = "saving the plot to file: \"" + savefig_kws["fname"] + "\"", end = "\n", pre = True )
        plt.savefig(**savefig_kws)
        self._progress.note( end = "\n", pre = True )

    ################################################################################################################################
    #### getLinSpace
    ################################################################################################################################

    def getLinSpace ( self
                    , skip = None
                    , lowerLim  : tp.Optional[ np.int32 ] = 1
                    , upperLim  : tp.Optional[ np.int32 ] = None
                    ):
        """

        Generate linearly-spaced **unique** integer numbers
        between the input lowerLim and upperLim. These numbers 
        can be used as the row indices in the plots.

        **Parameters**

            skip (optional)

                The linear spacing between the generated points.
                If ``skip`` is specified as input, any input value 
                for ``npoint`` will be ignored.
                The default value is ``None``.

            lowerLim (optional)

                The natural (non-logarithmic) lower limit of the
                generated linearly-spaced integer numbers.
                If not provided, the default value is ``1``.

            upperLim (optional)

                The natural (non-logarithmic) upper limit of the
                generated linearly-spaced integer numbers.
                If not provided, the default value is the maximum
                of the number of the rows of the input dataframe
                to the ``BasePlot`` constructor.

        **Returns**

            A set of unique linearly-spaced integer numbers.

        **Example**

            .. code-block:: python

                rows = getLinSpace(3, 1, 10000)

        """

        if upperLim is None and self._dfref is not None:
            upperLim = len(self._dfref().index)
        elif upperLim is None and self.rows is not None:
            upperLim = len(self.rows[-1])
        else:
            pm.abort( msg   = "Either the input argument ``upperLim`` or the dataFrame of the plot object" + newline
                            + "must be given in order to generate a linear range of values. " + newline
                            + "Here is the documentation of getLinSpace():" + newline
                            + newline
                            + self.getLinSpace.__doc__
                            + newline
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = self._methodName
                    )
        if skip is None: 
            skip = (upperLim - lowerLim) // 256
            if skip<1: skip = 1

        return range(lowerLim,upperLim,skip)

    ################################################################################################################################
    #### getLogLinSpace
    ################################################################################################################################

    def getLogLinSpace  ( self
                        , base      : tp.Optional[ np.float64 ] = 1.2
                        , logskip   : tp.Optional[ np.int32 ] = 0.2
                        , lowerLim  : tp.Optional[ np.int32 ] = 1
                        , upperLim  : tp.Optional[ np.int32 ] = None
                        ):
        """

        Generate logarithmically-uniformly-spaced **unique** integer 
        numbers between the input lowerLim and upperLim. These numbers 
        are to be used as the row indices in the plots.

        **Parameters**

            base (optional)

                The base of the logarithm used for generating the
                logarithmically-uniform range of integers.
                The default value is ``1.2``.

            logskip (optional)

                The minimum logarithmic space jump between
                the generated log-linearly-spaced integer numbers.
                The default value is ``0.2``.

            lowerLim (optional)

                The natural (non-logarithmic) lower limit of the
                generated log-linearly-spaced integer numbers.
                If not provided, the default value is ``1``.

            upperLim (optional)

                The natural (non-logarithmic) upper limit of the
                generated log-linearly-spaced integer numbers.
                If not provided, the default value is the maximum
                of the number of the rows of the input dataframe
                to the BasePlot constructor.

        **Returns**

            A set of unique log-linearly-spaced integer numbers.

        **Example**

            .. code-block:: python

                rows = getLogLinSpace(1.01, 1, 1, 10000)

        """

        if upperLim is None and self._dfref is not None:
            upperLim = len(self._dfref().index)
        elif upperLim is None and self.rows is not None:
            upperLim = len(self.rows[-1])
        else:
            pm.abort( msg   = "Either the input argument ``upperLim`` or the dataFrame of the plot object" + newline
                            + "must be given in order to generate a log-linear range of values. " + newline
                            + "Here is the documentation of getLogLinSpace():" + newline
                            + newline
                            + self.getLogLinSpace.__doc__
                            + newline
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = self._methodName
                    )

        return pm.utils.getLogIntSpace(base,logskip,lowerLim,upperLim)

    ################################################################################################################################
    #### _constructBasePlot
    ################################################################################################################################

    def _constructBasePlot(self):
        """

        Generate Figure and Axes instances if needed and return nothing.

            **Parameters**

                None

            **Returns**

                None

        """

        end = "\n" if "grid"==self._type.name else ""
        self._progress.note( msg = "making the " + self._type.name + " plot... ", end = end, pre = True )

        import seaborn as sns
        from matplotlib import pyplot as plt
        if self._type.is3d: from mpl_toolkits.mplot3d import Axes3D

        if self.set.enabled:
            sns.set( **vars(self.set.kws) )

        if not (self._type.isJointplot or self._type.isGridPlot): # jointplot and gridplot create their own figure and axes

            if self.figure.enabled:
                self.currentFig.figure = plt.figure( **vars(self.figure.kws) )
                if self._type.is3d:
                    self.currentFig.axes = Axes3D(self.currentFig.figure, **vars(self.axes3d.kws) )
                else:
                    self.currentFig.axes = self.currentFig.figure.gca( **vars(self.axes.kws) )
            else:
                self.currentFig.figure = plt.gcf() # self.currentFig.axes.get_figure()
                self.currentFig.axes = self.currentFig.figure.gca()

    ################################################################################################################################
    #### _finalizeBasePlot
    ################################################################################################################################

    def _finalizeBasePlot(self):
        """

        Set the base Figure properties, such as legends, ...

            **Parameters**

                None

            **Returns**

                None

        """

        ############################################################################################################################
        #### _finalizeBasePlot
        ############################################################################################################################

        if not (self._type.isGridPlot or self._type.isHeatmap or self._type.isDiffusionPlot) and self.legend.enabled:

            labelsNeeded = False
            if not hasattr(self.legend.kws, "labels"):
                labelsNeeded = True
            elif self.legend.kws.labels is None:
                labelsNeeded = True

            if labelsNeeded and hasattr(self.legend,"_labels") and self.legend._labels is not None:
                self.legend.kws.labels = self.legend._labels

            if self._type.isJointplot:
                self.currentFig.legend = self.currentFig.jointplot.ax_joint.legend( **vars(self.legend.kws) )
            else:
                self.currentFig.legend = self.currentFig.axes.legend( **vars(self.legend.kws) )

            self.currentFig.legend.set_zorder(20) # bring legend to the top

        if not self._type.is3d:
            import matplotlib.pyplot as plt
            plt.tight_layout()

        self._progress.note()

    ################################################################################################################################
    #### _checkDataType
    ################################################################################################################################

    def _checkDataType(self):
        """

        Verify the integrity of the dataFrame and return nothing.

            **Parameters**

                None

            **Returns**

                None

        """

        import pandas as pd

        fatalmsg = None
        if self._dfref is None:
            fatalmsg = "It appears that no data has been passed for plotting." + newline
        elif not isinstance(self._dfref,wref.ref):
            fatalmsg = "It appears that you have messed with the " + newline + " internal representation of data in the object." + newline
        elif not isinstance(self._dfref(), pd.DataFrame):
            fatalmsg = ""
        if fatalmsg is not None:
            raise Exception ( fatalmsg
                            + "The input data must be a pandas' dataframe." + newline
                            + "Please pass a dataFrame to the constructor or at" + newline
                            + "the time of calling the object (which is callable with" + newline
                            + "the same input arguments as the object's constructor."
                            )

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the BasePlot class:" + newline \
                    + newline \
                    + self.__doc__
        return docString

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import os
import sys
import typing as tp
import ctypes as ct

import _paramonte as pm
import _SpecBase as SpecBase
import _SpecMCMC as SpecMCMC
import _SpecDRAM as SpecDRAM
from _ReportFileContents import ReportFileContents
from _RestartFileContents import RestartFileContents
from _TabularFileContents import TabularFileContents

Struct = pm.Struct
newline = pm.newline

####################################################################################################################################
#### ParaMonteSampler class
####################################################################################################################################

class ParaMonteSampler:
    """

    This is the **ParaMonteSampler** base class for the ParaMonte
    sampler routines. This class is NOT meant to be directly accessed
    or called by the user of the ParaMonte library. However, its children,
    such the ParaDRAM sampler class will be directly accessible to the public.

        **Parameters**

            methodName

                A string representing the name of the ParaMonte sampler 
                that is to be instantiated.

        **Attributes**

            buildMode

                optional string argument with the default value "release".
                possible choices are:

                    "debug"

                        to be used for identifying sources of bug
                        and causes of code crash.

                    "release"

                        to be used in all other normal scenarios
                        for maximum runtime efficiency.

            mpiEnabled

                optional logical (boolean) indicator which is ``False`` by default.
                If it is set to ``True``, it will cause the ParaMonte simulation
                to run in parallel on the requested number of processors.
                See the class documentation guidelines in the above for
                information on how to run a simulation in parallel.

            reportEnabled

                optional logical (boolean) indicator which is ``True`` by default.
                If it is set to ``True``, it will cause extensive guidelines to be
                printed on the standard output as the simulation or post-processing
                continues with hints on the next possible steps that could be taken
                in the process. If you do not need such help and information set
                this variable to ``False`` to silence all output messages.

            inputFile

                optional string input representing the path to
                an external input namelist of simulation specifications.
                USE THIS OPTIONAL ARGUMENT WITH CAUTION AND
                ONLY IF YOU KNOW WHAT YOU ARE DOING.

                **WARNING**

                Specifying an input file will cause the sampler to ignore 
                all other simulation specifications set by the user via 
                sampler instance's `spec`-component attributes.

            spec

                A Python structure containing all simulation specifications.
                All simulation attributes are by default set to appropriate
                values at runtime. To override the default simulation
                specifications, set the ``spec`` attributes to some
                desired values of your choice.

                If you need help on any of the simulation specifications, try
                the supplied ``helpme()`` function in this component.

                If you wish to reset some specifications to the default values, 
                simply set them to ``None``.

        **Methods**

            See below for information on the methods.

        **Returns**

            Object of class ParaMonteSampler.

    """

    ################################################################################################################################
    #### ParaMonteSampler constructor
    ################################################################################################################################

    def __init__( self
                , methodName : str
                ):

        self._methodName = methodName
        self._libName = []
        self._ndim = []

        self._method = Struct()
        self._method.isParaDRAM = False
        self._method.isParaNest = False
        self._method.isParaTemp = False

        if self._methodName=="ParaDRAM":
            self._method.isParaDRAM = True
            self._objectName = "pmpd"
        else:
            pm.abort( msg   = "Internal error occurred. No sampling method other than ParaDRAM is currently " + newline
                            + "supported. Among its output files or simply, the path to the specific file " + newline
                            + "to be read. Please report this error at:" + newline
                            + newline
                            + "    " + pm.website.github.issues.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        self.buildMode = "release"
        self.mpiEnabled = False
        self.reportEnabled = not self.mpiEnabled
        self.inputFile = ""

        ############################################################################################################################
        #### ParaMonte specifications
        ############################################################################################################################

        self.spec = pm.utils.FrozenClass()

        # ParaMonte variables

        self.spec.sampleSize                               = None
        self.spec.randomSeed                               = None
        self.spec.description                              = None
        self.spec.outputFileName                           = None
        self.spec.outputDelimiter                          = None
        self.spec.chainFileFormat                          = None
        self.spec.variableNameList                         = None
        self.spec.restartFileFormat                        = None
        self.spec.outputColumnWidth                        = None
        self.spec.overwriteRequested                       = None
        self.spec.outputRealPrecision                      = None
        self.spec.silentModeRequested                      = None
        self.spec.domainLowerLimitVec                      = None
        self.spec.domainUpperLimitVec                      = None
        self.spec.parallelizationModel                     = None
        self.spec.progressReportPeriod                     = None
        self.spec.targetAcceptanceRate                     = None
        self.spec.mpiFinalizeRequested                     = None
        self.spec.maxNumDomainCheckToWarn                  = None
        self.spec.maxNumDomainCheckToStop                  = None

        if self._method.isParaDRAM:

            # ParaMCMC variables

            self.spec.chainSize                            = None
            self.spec.scaleFactor                          = None
            self.spec.startPointVec                        = None
            self.spec.proposalModel                        = None
            self.spec.proposalStartCovMat                  = None
            self.spec.proposalStartCorMat                  = None
            self.spec.proposalStartStdVec                  = None
            self.spec.sampleRefinementCount                = None
            self.spec.sampleRefinementMethod               = None
            self.spec.randomStartPointRequested            = None
            self.spec.randomStartPointDomainLowerLimitVec  = None
            self.spec.randomStartPointDomainUpperLimitVec  = None

            # ParaDRAM variables

            self.spec.adaptiveUpdateCount                  = None
            self.spec.adaptiveUpdatePeriod                 = None
            self.spec.greedyAdaptationCount                = None
            self.spec.delayedRejectionCount                = None
            self.spec.burninAdaptationMeasure              = None
            self.spec.delayedRejectionScaleFactorVec       = None

            self.spec.helpme = SpecDRAM.helpme

        self.spec._freeze()

    ################################################################################################################################
    #### _runSampler
    ################################################################################################################################

    def _runSampler ( self
                    , ndim          : int
                    , getLogFuncRaw : tp.Callable[[int,tp.List[float]], float]
                    , inputFile     : tp.Optional[str] = None
                    ) -> None:
        """

        Run ParaMonte sampler and return nothing. This method is identical to
        the ``runSampler()`` method, except that the input ``point`` parameter to
        the user-provided input objective function ``getLogFuncRaw(ndim,point)`` is
        a C-style raw pointer. This requires the user to guarantee that ``point`` will
        be always used with array bounds in their implementation of the objective function.
        The use of ``_runSampler()`` in place of ``runSampler()`` leads to a slight
        performance gain in the simulations.

            **Example serial usage**

            Copy and paste the following code enclosed between the
            two comment lines in your python/ipython/jupyter session
            (ensure the indentations of the pasted lines comply with Python rules):

            .. code-block:: python
                :linenos:

                ##################################
                import paramonte as pm
                import numpy as np
                def getLogFuncRaw(ndim,point):
                    # return the log of the standard multivariate
                    # Normal density function with ndim dimensions
                    return -0.5 * np.sum( np.double( point[0:ndim] )**2 )
                pmpd = pm.ParaDRAM()
                pmpd._runSampler( ndim = 4                      # length of point
                                , getLogFuncRaw = getLogFuncRaw # the objective function
                                )
                ##################################

            where,

                ndim

                    represents the number of dimensions of the domain of
                    the user's objective function ``getLogFuncRaw(ndim, point)``
                    and,

                getLogFuncRaw(ndim, point)

                    represents the user's objective function to be sampled,
                    where,

                        ndim

                            is a 32-bit integer, representing the number of
                            dimensions of the domain of the user-provided
                            objective function.

                        point

                            is a C-style array-pointer of length ``ndim``
                            and type float64. Note that the bounds of
                            ``point`` must be always specified wherever
                            it is used within the objective function.

                    On output, it must return the natural logarithm of
                    the objective function.

            **Parameters**

                All input parameters have the same meaning as the parameters
                of ``runSampler()``. The only difference is in the input
                parameters to the objective function ``getLogFuncRaw``.

            **Returns**

                None

        """

        #### verify ndim

        if not isinstance(ndim,int) or ndim<1:
            pm.abort( msg   = "The input argument ndim must be a positive integer," + newline
                            + "representing the number of dimensions of the domain of" + newline
                            + "the user's objective function ``getLogFunc()``." + newline
                            + "You have entered ndim = " + str(ndim)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        #### verify getLogFuncRaw

        if not callable(getLogFuncRaw):
            pm.abort( msg   = "The input argument ``getLogFuncRaw`` must be a callable function." + newline
                            + "It represents the user's objective function to be sampled," + newline
                            + "which must take an input integer ndim representing the number of" + newline
                            + "dimensions of the domain of the objective function to be samples and," + newline
                            + "a second input argument of type numpy float64 array of length ndim." + newline
                            + "On return it must return the natural logarithm of the objective function."
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        #### verify mpiEnabled

        if not isinstance(self.mpiEnabled,bool):
            pm.abort( msg   = "The sampler attribute ``mpiEnabled`` must be of type bool." + newline
                            + "It is an optional logical (boolean) indicator which is False by default." + newline
                            + "If it is set to True, it will cause the ParaMonte simulation" + newline
                            + "to run in parallel on the requested number of processors." + newline
                            + "See the ParaMonte class information on how to run a simulation " + newline
                            + "in parallel. You have entered mpiEnabled = " + str(self.mpiEnabled)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )
        if self.mpiEnabled: self.reportEnabled = False

        #### verify buildMode

        buildMode = "release" if self.buildMode is None else self.buildMode

        stype = None
        dummyList = None
        errorOccurred = True
        if isinstance(buildMode,str):
            errorOccurred = False
            if "-" in buildMode:
                dummyList = buildMode.split("-")
                buildMode = dummyList[0] # build type
                stype = dummyList[1] # compiler suite

        if not errorOccurred: errorOccurred = buildMode not in ["release", "testing", "debug"]
        if not errorOccurred and stype is not None: errorOccurred = not (stype=="gnu" or stype=="intel")
        if errorOccurred:
            if dummyList is not None: buildMode = "-".join(dummyList)
            pm.abort( msg   = "The object attribute ``buildMode`` must be of type ``str``." + newline
                            + "It is an optional string argument with default value \"release\"." + newline
                            + "possible choices are:" + newline
                            + newline
                            + "    \"debug\":" + newline
                            + newline
                            + "        to be used for identifying sources of bug" + newline
                            + "        and causes of code crash." + newline
                            + newline
                            + "    \"release\":" + newline
                            + newline
                            + "        to be used in all other normal scenarios" + newline
                            + "        for maximum runtime efficiency." + newline
                            + newline
                            + "You have entered buildMode = " + str(buildMode)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        #### verify inputFile

        if inputFile is not None and not isinstance(inputFile,str):
            pm.abort( msg   = "The input argument ``inputFile`` must be of type str." + newline
                            + "It is an optional string input representing the path to" + newline
                            + "an external input namelist of simulation specifications." + newline
                            + "USE THIS OPTIONAL ARGUMENT WITH CAUTION AND" + newline
                            + "ONLY IF YOU KNOW WHAT YOU ARE DOING." + newline
                            + "Specifying this option will cause the ParaMonte sampler " + newline
                            + "to ignore all other simulation specifications set by " + newline
                            + "the user via sampler's instance ``spec`` attributes." + newline
                            + "You have entered inputFile = " + str(inputFile)
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        inputFileVec_pntr, inputFileLen = self._getInputFile(inputFile)

        if self.mpiEnabled:
            parallelism = "_mpi"
        else:
            parallelism = ""
            pm.note( msg   = "Running the " + self._methodName + " sampler in serial mode..." + newline
                            + "To run the " + self._methodName + " sampler in parallel mode visit:" + newline
                            + newline
                            + "    " + pm.website.home.url + newline
                            + newline
                            + "If you are using Jupyter notebook, check the Jupyter's " + newline
                            + "terminal window for realtime simulation progress and report."
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        #if len(sys.argv)>1:
        #    if sys.argv[1]=="p":
        #        pm.note( msg = Running sampler in parallel mode...
        #                , methodName = self._methodName
        #                )
        #        print("\nRunning sampler in parallel mode..." + newline)
        #        libName += "_mpi"
        #else:
        #    print("\nRunning ParaMonte sampler in serial mode..." + newline)
        #try:
        #    from mpi4py import MPI
        #    comm = MPI.COMM_WORLD
        #    libName += "_mpi"
        #    if comm.size==1:
        #        print("\nRunning ParaMonte sampler in serial mode..." + newline)
        #        if MPI.Is_initialized():
        #            print("Hello")
        #            MPI.Finalize()
        #    elif comm.rank==0:
        #        print("\nRunning ParaMonte sampler in parallel mode on {} processes..." + newline.format(comm.size))
        #        comm.barrier()
        #except ImportError:
        #    print("\nImportError occurred..." + newline)
        #    print("\nRunning ParaMonte sampler in serial mode..." + newline)

        sys.stdout.flush()

        #### setup env

        if pm.platform.isWin32:

            if "PATH" in os.environ:
                os.environ["PATH"] = pm.path.root + os.pathsep + os.environ["PATH"]
            else:
                os.environ["PATH"] = pm.path.root

            #mpiFound = False
            pathList = os.environ["PATH"].split(";")
            for path in pathList:
                pathLower = path.lower().replace("\\","")
                if ("mpiintel64bin" in pathLower):
                    #mpiFound = True
                    #bldMode = buildMode
                    #if bldMode=="testing": bldMode = "release"
                    mpiPath = os.path.join(path,"release")
                    os.environ["PATH"] = mpiPath + os.pathsep + os.environ["PATH"]
                    libfabricPath = os.path.join(os.path.dirname(path),"libfabric","bin")
                    os.environ["PATH"] = libfabricPath + os.pathsep + os.environ["PATH"]
                    break

        else:

            if "LD_LIBRARY_PATH" not in os.environ:
                os.environ["LD_LIBRARY_PATH"] = "."
                if self.reportEnabled and pm.platform.isLinux:
                    pm.warn( msg   = "LD_LIBRARY_PATH environmental variable is not defined in " + newline
                                    + "your Python session. Consider running the following command " + newline
                                    + "in your Bash shell before reopening Python and using ParaMonte: " + newline
                                    + newline
                                    + "    export LD_LIBRARY_PATH=."
                            , methodName = self._methodName
                            , marginTop = 1
                            , marginBot = 1
                            )
            libdir = "/usr/lib"
            if os.path.isdir(libdir):
                os.environ["LD_LIBRARY_PATH"]  = libdir + os.pathsep + os.environ["LD_LIBRARY_PATH"]
            libdir = "/usr/local/lib"
            if os.path.isdir(libdir):
                os.environ["LD_LIBRARY_PATH"]  = libdir + os.pathsep + os.environ["LD_LIBRARY_PATH"]
            libdir = "/usr/lib64"
            if os.path.isdir(libdir):
                os.environ["LD_LIBRARY_PATH"]  = libdir + os.pathsep + os.environ["LD_LIBRARY_PATH"]
            libdir = "/usr/local/lib64"
            if os.path.isdir(libdir):
                os.environ["LD_LIBRARY_PATH"]  = libdir + os.pathsep + os.environ["LD_LIBRARY_PATH"]
            os.environ["LD_LIBRARY_PATH"]  = pm.path.root + os.pathsep + os.environ["LD_LIBRARY_PATH"]

            from _pmreqs import getLocalInstallDir
            localInstallDir = getLocalInstallDir()

            if localInstallDir.gnu.root is not None:
                for object in os.scandir(localInstallDir.gnu.root):
                    if object.is_dir() and ("lib" in object.name):
                        os.environ["LD_LIBRARY_PATH"] = object.path + os.pathsep + os.environ["LD_LIBRARY_PATH"]

            if localInstallDir.mpi.root is not None:
                if localInstallDir.mpi.bin is not None: os.environ["PATH"] = localInstallDir.mpi.bin + os.pathsep + os.environ["PATH"]
                for object in os.scandir(localInstallDir.mpi.root):
                    if object.is_dir() and ("lib" in object.name):
                        os.environ["LD_LIBRARY_PATH"] = object.path + os.pathsep + os.environ["LD_LIBRARY_PATH"]
                if localInstallDir.mpi.lib is not None: os.environ["LD_LIBRARY_PATH"] = localInstallDir.mpi.lib + os.pathsep + os.environ["LD_LIBRARY_PATH"]

        # import ParaMonte dll define result (None) AND argument (pointer to a c function) type

        buildModeList = ["release","testing","debug"]
        buildModeList.pop(buildModeList.index(buildMode))
        buildModeList.insert(0,buildMode)
        pmcsList = ["intel","gnu"]
        if stype is not None:
            pmcsList.pop(pmcsList.index(stype))
            pmcsList.insert(0,stype)

        libNameSuffix = parallelism +   { "windows" : ".dll"
                                        , "cygwin"  : ".dll"
                                        , "darwin"  : ".dylib"
                                        , "linux"   : ".so"
                                        }.get(pm.platform.osname, ".so")

        libPath = None
        libFound = False
        libNamePrefix = "libparamonte_python_" + pm.platform.osname.lower() + "_" + pm.platform.arch + "_"
        from ctypes.util import find_library
        for buildMode in buildModeList:

            for pmcs in pmcsList:

                libName = libNamePrefix + pmcs + "_" + buildMode + "_dynamic_heap" + libNameSuffix
                libPath = find_library(libName)
                if libPath is None: libPath = os.path.join( pm.path.root, libName )

                libFound = os.path.isfile(libPath)
                if libFound: break

            if libFound: # check if lib file exists
                break
            #else:
            #    if self.reportEnabled:
            #        pm.warn( msg   = "ParaMonte dynamic library for the requested build mode " + buildMode + " not found." + newline
            #                        + "Searching for ParaMonte dynamic library in other build modes..."
            #                , methodName = self._methodName
            #                , marginTop = 1
            #                , marginBot = 1
            #                )
            #    #libName = libName.replace(buildMode,mode)
            #    #buildMode = mode

        if not libFound:
            if self.mpiEnabled:
                parallelMsg = ("This happens frequently with parallel simulations and the " + newline
                            + "most likely reason is that the user did NOT carefully follow " + newline
                            + "the ParaMonte instructions to successfully install and define " + newline
                            + "the variables of the MPI runtime library on their system. " + newline
                            + "To learn these about these instructions, type the following " + newline
                            + "in your Python session, " + newline
                            + newline
                            + "    import paramonte as pm" + newline
                            + "    pm.verify()" + newline
                            + newline
                            + "Then, carefully follow the instructions provided to define " + newline
                            + "the MPI runtime variables in your current Python session. " + newline
                            + "If the error still persists, please report this issue at: " + newline
                            )
            else:
                parallelMsg = "Please report this issue at:" + newline
            if pm.platform.isWin32:
                from _pmreqs import buildInstructionNoteWindows
                buildMsg = buildInstructionNoteWindows
            else:
                from _pmreqs import buildInstructionNoteUnix
                buildMsg = newline + newline + buildInstructionNoteUnix
            pm.abort( msg   = "Exhausted all possible ParaMonte dynamic library search" + newline
                            + "names but could not find any compatible library." + newline
                            #+ "Last search:" + newline
                            #+ newline
                            #+ "    " + libPath + newline
                            #+ newline
                            + "It appears your ParaMonte library is missing some files. " + newline
                            + parallelMsg
                            + newline
                            + "    " + pm.website.github.issues.url + newline
                            + newline
                            + "Visit," + newline
                            + newline
                            + "    " + pm.website.home.url + newline
                            + newline
                            + "for instructions on how to build the ParaMonte library" + newline
                            + "object files on your system."
                            + buildMsg
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        # define ctypes wrapper function, with the proper result and argument types
        _getLogFuncRaw_proc = ct.CFUNCTYPE ( ct.c_double                  # function result
                                           #, ct.POINTER(ct.c_int32)      # ndim
                                            , ct.c_int32                   # ndim
                                            , ct.POINTER(ct.c_double)     # point
                                            )
        getLogFuncRaw_pntr = _getLogFuncRaw_proc(getLogFuncRaw)

        try:

            pmdll = ct.CDLL(libPath)

        except Exception as e:

            import logging
            logger = logging.Logger("catch_all")
            logger.error(e, exc_info=True)

            from _pmreqs import buildInstructionNote
            pm.abort( msg  = "Failed to load the required ParaMonte shared library." + newline
                            + "This is either due to the incompatibility of the DLL with your" + newline
                            + "platform or due to missing some required dependent libraries." + newline
                            + "In either case, you can likely resolve this error by building." + newline
                            + "the required ParaMonte shared libraries on your system." + newline
                            + newline
                            + "Visit," + newline
                            + newline
                            + "    " + pm.website.home.url + newline
                            + newline
                            + "for instructions to build the ParaMonte library on your system." + newline
                            + newline
                            + "Please report this issue at:" + newline
                            + newline
                            + "    " + pm.website.github.issues.url + newline
                            + newline
                            + buildInstructionNote
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        pmdll.runParaDRAM.restype = ct.c_int32
        #pmdll.runParaDRAM.restype = None
        #pmdll.runParaDRAM.argtypes =    [ ct.POINTER(ct.c_int32)     # ndim
        pmdll.runParaDRAM.argtypes =    [ ct.c_int32                   # ndim
                                        , _getLogFuncRaw_proc           # procedure
                                        , ct.POINTER(ct.c_char)       # inputFile byte object
                                        , ct.c_int32                   # lenInpuFile
                                       #, ct.POINTER(ct.c_size_t)     # lenInpuFile
                                        , ]

        #def getLogFuncRawWrapper(ndim_pntr,point): return getLogFuncRaw(ndim[0],point)

        # construct procedure pointer
        #def getLogFuncRawWrapper(ndim,point): return getLogFuncRaw(np.array(point[0:ndim]))
        #getLogFuncRaw_pntr = _getLogFuncRaw_proc(getLogFuncRawWrapper)

        # construct ndim pointer
        #ndim_pntr = ct.byref(ct.c_int32(ndim))

        # call ParaMonte
        #pmdll.runParaDRAM   ( ndim_pntr
        #pmdll.runParaDRAM   ( ct.c_int32(ndim)


        if self._method.isParaDRAM:

            errFlag = pmdll.runParaDRAM ( ct.c_int32(ndim)
                                        , getLogFuncRaw_pntr
                                        , inputFileVec_pntr
                                        , ct.c_int32(inputFileLen)
                                       #, inputFileLen_pntr
                                        )

            if errFlag!=0:

                # first check for old existing files:

                existingFileList = []
                outputFileName = os.path.abspath(self.spec.outputFileName)
                if not os.path.isdir(outputFileName):
                    import glob
                    existingFileList = glob.glob(outputFileName+"*")

                if self._method.isParaDRAM and len(existingFileList)>1:
                    existingSimulationMsg = ( "It appears that an old " + self._methodName + " simulation with " + newline
                                            + "the same output file names exists in the specified path:" + newline
                                            + newline
                                            + newline.join(existingFileList)
                                            + newline + newline
                                            + "Keep in mind that old simulation files CANNOT BE OVERWRITTEN " + newline
                                            + "and attempting to do so will cause the new simulation to crash. " + newline
                                            + "If this is the case, specify a new filename prefix for the simulation " + newline
                                            + "output files or set it to `None` in the simulation's input specifications " + newline
                                            + "so that the sampler can automatically generate unique output " + newline
                                            + "filenames for your simulation."
                                            )
                else:
                    existingSimulationMsg = ""

                pm.abort( msg   = "The simulation failed. For more information, checkout the " + newline
                                + "simulation output error message on your your Bash / Python " + newline
                                + "terminal or command-prompt, also at the end of the output report " + newline
                                + "file, if it has been generated:" + newline
                                + newline
                                + "    " + outputFileName + "_report.txt" + newline
                                + newline
                                + existingSimulationMsg
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 1
                        )

        #def isLoaded(libPath):
        #    abslibPath =
        #    return os.system("lsof -p {} | grep {} > /dev/null".format( os.getpid(), os.path.abspath(libPath) )) == 0
        #def dlclose(libdll): libdll.dlclose(libdll._handle)

        if pm.platform.isWin32:
            handle = ct.windll.kernel32.LoadLibraryA(libPath)
            ct.windll.kernel32.FreeLibrary(handle)
        else:
           #while isLoaded(libPath):
           #    dlclose(pmdll._handle)
            try:
                ct.dlclose(pmdll._handle)
            except:
                if self.reportEnabled:
                    pm.warn ( msg   = "Failed to properly close the ParaMonte shared library file. " + newline
                                    + "This should not cause any major problems, unless you intend to " + newline
                                    + "run a new ParaMonte simulation, in which case, you may want to " + newline
                                    + "exit and re-enter your Python environment."
                            , methodName = self._methodName
                            , marginTop = 1
                            , marginBot = 1
                            )

        if self.reportEnabled:
            pm.note( msg    = "To read the generated output files, try:" + newline
                            + newline
                            + "    " + self._objectName + ".readReport()      # to read the summary report from the output report file." + newline
                            + "    " + self._objectName + ".readSample()      # to read the final i.i.d. sample from the output sample file." + newline
                            + "    " + self._objectName + ".readChain()       # to read the uniquely-accepted points from the output chain file." + newline
                            + "    " + self._objectName + ".readMarkovChain() # to read the Markov Chain. NOT recommended for very large chains." + newline
                            + "    " + self._objectName + ".readRestart()     # to read the contents of an ASCII-format output restart file." + newline
                            + "    " + self._objectName + ".readProgress()    # to read the contents of an output progress file." + newline
                            + newline
                            + "where you should replace `" + self._objectName + "` with your " + self._methodName + " sampler's object name." + newline
                            + "For more information and examples on the usage, visit:" + newline
                            + newline
                            + "    " + pm.website.home.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        return None

    ################################################################################################################################
    #### _getInputFile()
    ################################################################################################################################

    def _getInputFile(self, inputFile):

        if inputFile is None:

            ########################################################################################################################
            #### begin namelist generation from arguments
            ########################################################################################################################

            nameList = ""

            # setup outputFileName if it is None

            if self.spec.outputFileName is None:
                self.spec.outputFileName = os.path.join( os.getcwd() , SpecBase.genOutputFileName(self._methodName) )
            else:
                if self.spec.outputFileName[-1] == "\\" or self.spec.outputFileName[-1] == "/":
                    self.spec.outputFileName = os.path.join ( os.path.abspath( self.spec.outputFileName ) , SpecBase.genOutputFileName(self._methodName) )

            # ParaMonte variables

            if   self.spec.sampleSize                               is not None: nameList += SpecBase.sampleSize                            (self.spec.sampleSize                         )
            if   self.spec.randomSeed                               is not None: nameList += SpecBase.randomSeed                            (self.spec.randomSeed                         )
            if   self.spec.description                              is not None: nameList += SpecBase.description                           (self.spec.description                        )
            if   self.spec.outputFileName                           is not None: nameList += SpecBase.outputFileName                        (self.spec.outputFileName                     )
            if   self.spec.outputDelimiter                          is not None: nameList += SpecBase.outputDelimiter                       (self.spec.outputDelimiter                    )
            if   self.spec.chainFileFormat                          is not None: nameList += SpecBase.chainFileFormat                       (self.spec.chainFileFormat                    )
            if   self.spec.variableNameList                         is not None: nameList += SpecBase.variableNameList                      (self.spec.variableNameList                   )
            if   self.spec.restartFileFormat                        is not None: nameList += SpecBase.restartFileFormat                     (self.spec.restartFileFormat                  )
            if   self.spec.outputColumnWidth                        is not None: nameList += SpecBase.outputColumnWidth                     (self.spec.outputColumnWidth                  )
            if   self.spec.overwriteRequested                       is not None: nameList += SpecBase.overwriteRequested                    (self.spec.overwriteRequested                 )
            if   self.spec.outputRealPrecision                      is not None: nameList += SpecBase.outputRealPrecision                   (self.spec.outputRealPrecision                )
            if   self.spec.silentModeRequested                      is not None: nameList += SpecBase.silentModeRequested                   (self.spec.silentModeRequested                )
            if   self.spec.domainLowerLimitVec                      is not None: nameList += SpecBase.domainLowerLimitVec                   (self.spec.domainLowerLimitVec                )
            if   self.spec.domainUpperLimitVec                      is not None: nameList += SpecBase.domainUpperLimitVec                   (self.spec.domainUpperLimitVec                )
            if   self.spec.parallelizationModel                     is not None: nameList += SpecBase.parallelizationModel                  (self.spec.parallelizationModel               )
            if   self.spec.progressReportPeriod                     is not None: nameList += SpecBase.progressReportPeriod                  (self.spec.progressReportPeriod               )
            if   self.spec.targetAcceptanceRate                     is not None: nameList += SpecBase.targetAcceptanceRate                  (self.spec.targetAcceptanceRate               )
            if   self.spec.mpiFinalizeRequested                     is not None: nameList += SpecBase.mpiFinalizeRequested                  (self.spec.mpiFinalizeRequested               )
            if   self.spec.maxNumDomainCheckToWarn                  is not None: nameList += SpecBase.maxNumDomainCheckToWarn               (self.spec.maxNumDomainCheckToWarn            )
            if   self.spec.maxNumDomainCheckToStop                  is not None: nameList += SpecBase.maxNumDomainCheckToStop               (self.spec.maxNumDomainCheckToStop            )

            if self._method.isParaDRAM:

                # ParaMCMC variables

                if   self.spec.chainSize                            is not None: nameList += SpecMCMC.chainSize                             (self.spec.chainSize                          )
                if   self.spec.scaleFactor                          is not None: nameList += SpecMCMC.scaleFactor                           (self.spec.scaleFactor                        )
                if   self.spec.startPointVec                        is not None: nameList += SpecMCMC.startPointVec                         (self.spec.startPointVec                      )
                if   self.spec.proposalModel                        is not None: nameList += SpecMCMC.proposalModel                         (self.spec.proposalModel                      )
                if   self.spec.proposalStartCovMat                  is not None: nameList += SpecMCMC.proposalStartCovMat                   (self.spec.proposalStartCovMat                )
                if   self.spec.proposalStartCorMat                  is not None: nameList += SpecMCMC.proposalStartCorMat                   (self.spec.proposalStartCorMat                )
                if   self.spec.proposalStartStdVec                  is not None: nameList += SpecMCMC.proposalStartStdVec                   (self.spec.proposalStartStdVec                )
                if   self.spec.sampleRefinementCount                is not None: nameList += SpecMCMC.sampleRefinementCount                 (self.spec.sampleRefinementCount              )
                if   self.spec.sampleRefinementMethod               is not None: nameList += SpecMCMC.sampleRefinementMethod                (self.spec.sampleRefinementMethod             )
                if   self.spec.randomStartPointRequested            is not None: nameList += SpecMCMC.randomStartPointRequested             (self.spec.randomStartPointRequested          )
                if   self.spec.randomStartPointDomainLowerLimitVec  is not None: nameList += SpecMCMC.randomStartPointDomainLowerLimitVec   (self.spec.randomStartPointDomainLowerLimitVec)
                if   self.spec.randomStartPointDomainUpperLimitVec  is not None: nameList += SpecMCMC.randomStartPointDomainUpperLimitVec   (self.spec.randomStartPointDomainUpperLimitVec)

                # ParaDRAM variables

                if   self.spec.adaptiveUpdateCount                  is not None: nameList += SpecDRAM.adaptiveUpdateCount                   (self.spec.adaptiveUpdateCount                )
                if   self.spec.adaptiveUpdatePeriod                 is not None: nameList += SpecDRAM.adaptiveUpdatePeriod                  (self.spec.adaptiveUpdatePeriod               )
                if   self.spec.greedyAdaptationCount                is not None: nameList += SpecDRAM.greedyAdaptationCount                 (self.spec.greedyAdaptationCount              )
                if   self.spec.delayedRejectionCount                is not None: nameList += SpecDRAM.delayedRejectionCount                 (self.spec.delayedRejectionCount              )
                if   self.spec.burninAdaptationMeasure              is not None: nameList += SpecDRAM.burninAdaptationMeasure               (self.spec.burninAdaptationMeasure            )
                if   self.spec.delayedRejectionScaleFactorVec       is not None: nameList += SpecDRAM.delayedRejectionScaleFactorVec        (self.spec.delayedRejectionScaleFactorVec     )

            nameList = "&" + self._methodName + " " + nameList + SpecBase.interfaceType() + SpecBase.systemInfoFilePath(pm.platform.systemInfoFilePath) + "/"

            ############################################################################################################################
            #### end namelist generation from arguments
            ############################################################################################################################

            inputFileVec_pntr = nameList.encode("utf-8")                # create byte-object from the internal input file

        else:

            if not self.mpiEnabled:
                pm.warn ( msg   = "Input namelist file is given by the user. " + newline
                                + "All simulation specifications will be read from the input file."
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 1
                        )

            inputFileVec_pntr = inputFile.encode("utf-8")                       # create byte-object from the external input file

        inputFileLen = len(inputFileVec_pntr) # byte-object length
        #inputFileLen_pntr = ct.byref( ct.c_size_t( len(inputFileVec_pntr) ) ) # pointer to byte-object length
        inputFileVec_pntr = ct.c_char_p( inputFileVec_pntr )                   # pointer to byte-object

        return inputFileVec_pntr, inputFileLen #_pntr

    ################################################################################################################################
    #### _setFileToRead()
    ################################################################################################################################

    def _setFileToRead(self, file, fileType, fileSuffix):
        if self.spec.outputFileName is None:
            file = os.getcwd()
            if self.reportEnabled:
                pm.warn ( msg   = "The ``file`` is neither given as input to ``read" + fileType.capitalize() + "()``" + newline
                                + "nor set as a simulation specification of the " + self._methodName + " object. " + newline
                                + "This information is essential, otherwise how could the output files be found?" + newline
                                + "All that is needed is the unique name (including path) of the simulation name " + newline
                                + "shared among its output files or simply, the path to the specific " + fileSuffix + newline
                                + " file to be read. For now, the " + self._methodName + " sampler will search " + newline
                                + "the current working directory for simulation output files that match the " + newline
                                + "filename pattern of " + fileSuffix + " files."
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 1
                        )
        else:
            file = self.spec.outputFileName
        return file

    ################################################################################################################################
    #### _setDelimiterToRead()
    ################################################################################################################################

    def _setDelimiterToRead(self, delimiter, fileType, fileSuffix):
        if self.spec.outputDelimiter is None:
            delimiter   = ","
            if self.reportEnabled:
                pm.warn ( msg   = "The ``delimiter`` is neither given as input to ``read" + fileType.capitalize() + "()``" + newline
                                + "nor set as a simulation specification of the " + self._methodName + " object. " + newline
                                + "This information is essential, otherwise how could the output files be parsed?" + newline
                                + "For now, the " + self._methodName + " sampler will assume a comma-separated " + newline
                                + "file format for the contents of the " + fileSuffix + " file(s) to be parsed."
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 1
                        )
        else:
            delimiter = self.spec.outputDelimiter
        return delimiter

    ################################################################################################################################
    #### readTabular()
    ################################################################################################################################

    def _readTabular( self
                    , file          : str
                    , fileType      : str
                    , delimiter     : str
                    , parseContents : bool
                    , renabled      : bool
                    ) -> tp.List[TabularFileContents] :
        """

        Read the contents of the file(s) whose path is given by the input argument ``file``.
        This function is not to be directly accessible to and callable by the users
        of the ParaMonte library.

            **Parameters**

                file

                    A string representing the path to the tabular file with
                    the default value of ``None``.

                    The path only needs to uniquely identify the simulation
                    to which the tabular file belongs. For example, specifying
                    ``"./mydir/mysim"`` as input will lead to a search for a 
                    file that begins with ``"mysim"`` and ends with the tabular file 
                    name's prefix, such as, ``"_sample.txt"``, inside the directory 
                    ``"./mydir/"``. If there are multiple files with such name, 
                    then all of them will be read and returned as a list.

                    The path can be also a world wide web address.

                    If this input argument is not provided by the user, the
                    value of the object attribute ``outputFileName`` will be 
                    used instead. At least one of the two mentioned routes 
                    must provide the path to the tabular file otherwise,
                    this method will break by calling ``sys.exit()``.

                fileType

                    A string containing the type of the file to be parsed.
                    Current options include but are not limited to:
                    ``sample``, ``chain``, ``markovChain``, ``progress``

                delimiter

                    An input string representing the delimiter used in the
                    output tabular file. If it is not provided as input argument,
                    the value of the corresponding object attribute outputDelimiter
                    will be used instead. If none of the two are available,
                    the default comma delimiter "," will be assumed and used.

                parseContents

                    If set to True, the contents of the file will be parsed and
                    stored in a component of the object named ``contents``.
                    The default value is ``True``.

                renabled

                    If set to ``False``, the contents of the file(s) will be 
                    stored as a list in a (new) component of the object with a
                    name that ends with the prefix ``List``. Otherwise, ``None`` 
                    will be the return value of the method. If set to ``True``, 
                    the reverse will done. The default value is ``False``.

            **Returns**

                List

                    A Python list of ``TabularFileContents`` objects, each 
                    of which corresponds to the contents of a unique restart 
                    file. The contents of each object is dependent on the 
                    type of the file that has been parsed.

        """

        if fileType=="sample":
            fileSuffix = "sample"
        elif fileType=="chain" or fileType=="markovChain":
            fileSuffix = "chain"
        elif fileType=="progress":
            fileSuffix = "progress"
        else:
            fileSuffix = None
            pm.abort( msg   = "Internal error occurred. The input fileType is not recognized." + newline
                            + "Please report this error at:" + newline
                            + newline
                            + "    " + pm.website.github.issues.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if file is None: file = self._setFileToRead(file, fileType, fileSuffix)
        if delimiter is None: delimiter = self._setDelimiterToRead(delimiter, fileType, fileSuffix)

        FileList, iswebfile = pm.utils.getFileList(file, fileSuffix, self._methodName, self.reportEnabled)

        tabularContentsList = []
        for file in FileList:

            file = os.path.abspath(file)
            if self.reportEnabled:
                pm.note ( msg = "processing " + fileSuffix + " file: " + file
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 0
                        )

            tabularContents = TabularFileContents   ( file = file
                                                    , fileType = fileType
                                                    , delimiter = delimiter
                                                    , parseContents = parseContents
                                                    , reportEnabled = self.reportEnabled
                                                    , methodName = self._methodName
                                                    )
            tabularContentsList.append(tabularContents)

        outputListName = fileType + "List"
        if renabled:
            outputListFullName = outputListName
            msg=( "The processed " + fileType + " files are now stored in the output variable as a " + newline
                + "Python list. For example, to access the contents of the first (or the only) " + newline
                + fileType + " file stored in an output variable named " + outputListFullName + ", try:"
                )
        else:
            outputListFullName = self._objectName + "." + outputListName
            setattr(self, outputListName, tabularContentsList)
            msg=( "The processed " + fileType + " files are now stored in the newly-created" + newline
                + "component `" + outputListName + "` of the " + self._methodName + " object as a Python list." + newline
                + "For example, to access the contents of the first (or the only) " + fileType + " file, try:"
                )

        if self.reportEnabled:

            if fileSuffix=="progress":
                specials = ""
            else:
                specials =  ( "    " + outputListFullName + "[0].plot.line3()        # to make 3D line plots." + newline
                            + "    " + outputListFullName + "[0].plot.scatter3()     # to make 3D scatter plots." + newline
                            + "    " + outputListFullName + "[0].plot.lineScatter3() # to make 3D line-scatter plots." + newline
                            + "    " + outputListFullName + "[0].plot.contour()      # to make fast 2D kernel density plots." + newline
                            + "    " + outputListFullName + "[0].plot.contourf()     # to make fast 2D kernel density filled contour plots." + newline
                            + "    " + outputListFullName + "[0].plot.contour3()     # to make fast 3D kernel density contour plots." + newline
                            + "    " + outputListFullName + "[0].plot.histplot()     # to make seaborn 1D distribution plots." + newline
                            + "    " + outputListFullName + "[0].plot.kdeplot1()     # to make seaborn 1D kernel density plots." + newline
                            + "    " + outputListFullName + "[0].plot.kdeplot2()     # to make seaborn 2D kernel density plots." + newline
                            + "    " + outputListFullName + "[0].plot.grid()         # to make GridPlot" + newline
                            )

            pm.note ( msg   = msg + newline
                            + newline
                            + "    " + outputListFullName + "[0].df" + newline
                            + newline
                            + "where you will have to replace `pmpd` with your " + self._methodName + " instance name." + newline
                            + "To access the plotting tools, try:" + newline
                            + newline
                            + "    " + outputListFullName + "[0].plot.<PRESS TAB TO SEE THE LIST OF PLOTS>" + newline
                            + newline
                            + "For example," + newline
                            + newline
                            + "    " + outputListFullName + "[0].plot.line()         # to make 2D line plots." + newline
                            + "    " + outputListFullName + "[0].plot.scatter()      # to make 2D scatter plots." + newline
                            + "    " + outputListFullName + "[0].plot.lineScatter()  # to make 2D line-scatter plots." + newline
                            + specials
                            + newline
                            + "To plot or inspect the variable autocorrelations or the correlation/covariance matrices, try:" + newline
                            + newline
                            + "    " + outputListFullName + "[0].stats.<PRESS TAB TO SEE THE LIST OF COMPONENTS>" + newline
                            + newline
                            + "For more information and examples on the usage, visit:" + newline
                            + newline
                            + "    " + pm.website.home.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if iswebfile: pm.utils.delFile( file = FileList[0], desc = "the temporarily-downloaded " + fileType + " file", methodName = self._methodName )

        if renabled:
            return tabularContentsList
        else:
            return None

    ################################################################################################################################
    #### readSample
    ################################################################################################################################

    def readSample  ( self
                    , file          : tp.Optional[str]  = None
                    , delimiter     : tp.Optional[str]  = None
                    , parseContents : tp.Optional[bool] = True
                    , renabled      : tp.Optional[bool] = False
                    ) -> tp.List[TabularFileContents] :
        """

        Return a list of the contents of a set of ParaDRAM output
        sample files whose names contain the user-provided input file.
        This method is to be only used for postprocessing of the output
        sample file(s) of an already finished ParaDRAM simulation.
        It is not meant to be called by all processes in parallel mode,
        although it is possible.

            **Parameters**

                file (optional)

                    A string representing the path to the output file with
                    the default value of ``None``.

                    The path only needs to uniquely identify the simulation
                    to which the output file belongs. For example, specifying
                    ``"./mydir/mysim"`` as input will lead to a search for a 
                    file that begins with ``"mysim"`` and ends with ``"_sample.txt"``
                    inside the directory ``"./mydir/"``. If there are multiple
                    files with such name, then all of them will be read
                    and returned as a list.

                    The path can be also a world wide web address.

                    If this input argument is not provided by the user, the
                    value of the object attribute outputFileName
                    will be used instead. At least one of the two mentioned
                    routes must provide the path to the output file otherwise,
                    this method will break by calling ``sys.exit()``.

                delimiter (optional)

                    Optional input string representing the delimiter used in the
                    output output file. If it is not provided as input argument,
                    the value of the corresponding output object's attribute 
                    ``outputDelimiter`` will be used instead. If none of the two 
                    are available, the default comma delimiter ``","`` will be 
                    assumed and used. 

                parseContents (optional)

                    If set to True, the contents of the file will be parsed and
                    stored in a component of the object named ``contents``.
                    The default value is ``True``.

                renabled (optional)

                    If set to ``False``, the contents of the file(s) will be 
                    stored as a list in a (new) component of the ParaDRAM object 
                    named ``sampleList`` and ``None`` will be the return value 
                    of the method. If set to ``True``, the reverse will done.
                    The default value is ``False``.

            **Returns**

                sampleList (optional)

                    A Python list of ``TabularFileContents`` objects, each 
                    of which corresponds to the contents of a unique restart 
                    file. Each object has the following components:

                        file

                            The full absolute path to the output file.

                        delimiter

                            The delimiter used in the output file.

                        ndim

                            The number of dimensions of the domain of the objective 
                            function from which the output has been drawn.

                        count

                            The number of sampled points in the output file.

                        plot

                            A structure containing the graphics tools for the 
                            visualization of the contents of the file.

                        df

                            The contents of the output file in the form of
                            a pandas-library DataFrame (hence called ``df``).

                        contents

                            corresponding to each column in the progress file, a property
                            with the same name as the column header is also created
                            for the object which contains the data stored in that column
                            of the progress file. These properties are all stored in the 
                            attribute ``contents``.

                If ``renabled = True``, the list of objects will be returned as the
                return value of the method. Otherwise, the list will be stored in a
                component of the ParaDRAM object named ``sampleList``.

        """

        return self._readTabular( file = file
                                , fileType = "sample"
                                , delimiter = delimiter
                                , parseContents = parseContents
                                , renabled = renabled
                                )

    ################################################################################################################################
    #### readChain
    ################################################################################################################################

    def readChain   ( self
                    , file          : tp.Optional[str] = None
                    , delimiter     : tp.Optional[str] = None
                    , parseContents : tp.Optional[bool] = True
                    , renabled      : tp.Optional[bool] = False
                    ) -> tp.List[TabularFileContents] :
        """

        Return a list of the contents of a set of ParaDRAM output
        chain files whose names begin the user-provided input file.
        This method is to be only used for postprocessing of the output
        chain file(s) of an already finished ParaDRAM simulation.
        It is not meant to be called by all processes in parallel mode,
        although it is possible.

            **Parameters**

                file (optional)

                    A string representing the path to the output file with
                    the default value of ``None``.

                    The path only needs to uniquely identify the simulation
                    to which the output file belongs. For example, specifying
                    ``"./mydir/mysim"`` as input will lead to a search for a file
                    that begins with ``"mysim"`` and ends with ``"_chain.txt"``
                    inside the directory ``"./mydir/"``. If there are multiple
                    files with such name, then all of them will be read
                    and returned as a list.

                    The path can be also a world wide web address.

                    If this input argument is not provided by the user, the
                    value of the object attribute outputFileName
                    will be used instead. At least one of the two mentioned
                    routes must provide the path to the output file otherwise,
                    this method will break by calling ``sys.exit()``.

                delimiter (optional)

                    Optional input string representing the delimiter used in the
                    output output file. If it is not provided as input argument,
                    the value of the corresponding output object's attribute 
                    ``outputDelimiter`` will be used instead. If none of the two 
                    are available, the default comma delimiter ``","`` will be 
                    assumed and used. 

                parseContents (optional)

                    If set to True, the contents of the file will be parsed and
                    stored in a component of the object named ``contents``.
                    The default value is ``True``.

                renabled (optional)

                    If set to ``False``, the contents of the file(s) will be 
                    stored as a list in a (new) component of the ParaDRAM object 
                    named ``chainList`` and ``None`` will be the return value 
                    of the method. If set to ``True``, the reverse will done.
                    The default value is ``False``.

            **Returns**

                chainList (optional)

                    A Python list of ``TabularFileContents`` objects, each 
                    of which corresponds to the contents of a unique restart 
                    file. Each object has the following components:

                        file

                            The full absolute path to the output file.

                        delimiter

                            The delimiter used in the output file.

                        ndim

                            The number of dimensions of the domain of the objective 
                            function from which the output has been drawn.

                        count

                            The number of sampled points in the output file.

                        plot

                            A structure containing the graphics tools for the 
                            visualization of the contents of the file.

                        df

                            The contents of the output file in the form of
                            a pandas-library DataFrame (hence called ``df``).

                        contents

                            corresponding to each column in the progress file, a property
                            with the same name as the column header is also created
                            for the object which contains the data stored in that column
                            of the progress file. These properties are all stored in the 
                            attribute ``contents``.

                If ``renabled = True``, the list of objects will be returned as the
                return value of the method. Otherwise, the list will be stored in a
                component of the ParaDRAM object named ``sampleList``.

        """

        return self._readTabular( file = file
                                , fileType = "chain"
                                , delimiter = delimiter
                                , parseContents = parseContents
                                , renabled = renabled
                                )

    ################################################################################################################################
    #### readProgress
    ################################################################################################################################

    def readProgress( self
                    , file          : tp.Optional[str] = None
                    , delimiter     : tp.Optional[str] = None
                    , parseContents : tp.Optional[bool] = True
                    , renabled      : tp.Optional[bool] = False
                    ) -> tp.List[TabularFileContents] :
        """

        Return a list of the contents of a set of ParaMonte output
        progress files whose names begin the user-provided input file.
        This method is to be only used for postprocessing of the output
        progress file(s) of an already finished ParaMonte simulation.
        It is not meant to be called by all processes in parallel mode,
        although it is possible.

            **Parameters**

                file (optional)

                    A string representing the path to the output file with
                    the default value of ``None``.

                    The path only needs to uniquely identify the simulation
                    to which the output file belongs. For example, specifying
                    ``"./mydir/mysim"`` as input will lead to a search for a file
                    that begins with ``"mysim"`` and ends with ``"_progress.txt"``
                    inside the directory ``"./mydir/"``. If there are multiple
                    files with such name, then all of them will be read
                    and returned as a list.

                    The path can be also a world wide web address.

                    If this input argument is not provided by the user, the
                    value of the object attribute outputFileName
                    will be used instead. At least one of the two mentioned
                    routes must provide the path to the progress file otherwise,
                    this method will break by calling ``sys.exit()``.

                delimiter (optional)

                    Optional input string representing the delimiter used in the
                    output progress file. If it is not provided as input argument,
                    the value of the corresponding object attribute outputDelimiter
                    will be used instead. If none of the two are available,
                    the default comma delimiter ``","`` will be assumed and used.

                parseContents (optional)

                    If set to True, the contents of the file will be parsed and
                    stored in a component of the object named ``contents``.
                    The default value is ``True``.

                renabled (optional)

                    If set to False, the contents of the file(s) will be stored
                    as a list in a (new) component of the sampler object named
                    ``progressList`` and ``None`` will be the return value of the
                    method. If set to ``True``, the reverse will be done.
                    The default value is ``False``.

            **Returns**

                A list of objects, each of which has the following properties:

                    file

                        The full absolute path to the file.

                    delimiter

                        The delimiter used in the file.

                    ncol

                        The number of columns of the file.

                    plot

                        A structure containing the graphics tools for the 
                        visualization of the contents of the file.

                    df

                        the contents of the progress file in the form of
                        a pandas-library DataFrame (hence called ``df``).

                    contents

                        corresponding to each column in the progress file, a property
                        with the same name as the column header is also created
                        for the object which contains the data stored in that column
                        of the progress file. These properties are all stored in the 
                        attribute ``contents``.

                If ``renabled = True``, the list of objects will be returned as the
                return value of the method. Otherwise, the list will be stored in a
                component of the sampler object named ``progressList``.

        """

        return self._readTabular( file = file
                                , fileType = "progress"
                                , delimiter = delimiter
                                , parseContents = parseContents
                                , renabled = renabled
                                )

    ################################################################################################################################
    #### readRestart
    ################################################################################################################################

    def readRestart ( self
                    , file          : tp.Optional[str] = None
                    , renabled      : tp.Optional[bool] = False
                    ) -> tp.List[RestartFileContents] :
        """

        Return a list of the contents of a set of the simulation(s) 
        output restart files whose names begin the user-provided input 
        file prefix, or as specified by the input simulation specification 
        ``SAMPLER.spec.outputFileName``, where SAMPLER can be an instance 
        of any one of the ParaMonte's sampler classes, such as ``ParaDRAM()``.

            **NOTE**

            Only restart output files in **ASCII format** can be read 
            via this method. The binary restart files are NOT meant to 
            be parsed via this method. To request for ASCII restart 
            output files in simulations, set the input simulation 
            specification

            .. code-block:: python

                SAMPLER.spec.restartFileFormat = "ascii",

            where ``SAMPLER`` can be an instance of any one of the 
            ParaMonte's sampler classes, such as ``ParaDRAM()``.

            **WARNING**
            
            Avoid using this routine for very large long simulations.
            Reading the full restart file of a large-scale simulation 
            problem can be extremely memory-intensive.

            **WARNING**

            This method is to be only used for post-processing of the 
            output restart file(s) of an already finished simulation. 
            It is NOT meant to be called by all processes in parallel 
            mode, although it is possible.

            **Parameters**

                file (optional)

                    A string representing the path to the restart file with the
                    default value of []. The path only needs to uniquely identify
                    the name of the simulation to which the restart file belongs.
                    For example, specifying ``"./mydir/mysim"`` as input will lead 
                    to a search for a file that begins with ``"mysim"`` and ends with
                    ``"_restart.txt"`` inside the directory ``"./mydir/"``.
                    If there are multiple files with such name, then all of them
                    will be read and returned as a list.
                    If this input argument is not provided by the user, the
                    value of the object's ``spec`` attribute ``outputFileName``
                    will be used instead.

                    **WARNING**

                    At least one of the two mentioned routes must
                    provide the path to the restart file. Otherwise,
                    this method will abort the program.

                    Example usage:

                        .. code-block:: python

                            pmpd.readRestart("./out/test_run_")

                        or,

                        .. code-block:: python

                            pmpd.spec.outputFileName = "./out/test_run_"
                            pmpd.readRestart()

                   Both of the above examples are equivalent.
                   The latter is recommended as it is less confusing.

                renabled (optional)

                    If set to ``False``, the contents of the file(s) will be 
                    stored as a list in a (new) component of the sampler object 
                    named ``restartList`` and ``None`` will be the return value 
                    of the method. If set to ``True``, the reverse will done.
                    The default value is ``False``.

            **Returns**

                restartList (optional)

                    A Python list of ``RestartFileContents`` objects, each 
                    of which corresponds to the contents of a unique restart 
                    file. Each object has the following components:

                        file

                            The full absolute path to the restart file.

                        ndim

                            The number of dimensions of the domain of the objective 
                            function for which the restart file was generated.

                        count

                            The number of restart writes to the file.

                        plot

                            A structure containing the graphics tools for the 
                            visualization of the contents of the file.

                        df

                            The contents of the restart file in the form of a
                            ``panda``'s dataframe (``df`` stands for **DataFrame**).

                        contents

                            A structure whose components contain the information 
                            retrieved about each of the entities in the file.

                        propNameList

                            A list of entities names parsed from the restart file.

                If no output argument is provided, a ``restartList`` property will be
                added to the parent sampler-object to which the method ``readRestart()``
                belongs.

        """

        fileType = "restart"
        fileSuffix = "restart"

        if file is None: file = self._setFileToRead(file, fileType, fileSuffix)

        FileList, iswebfile = pm.utils.getFileList(file, fileSuffix, self._methodName, self.reportEnabled)

        restartContentsList = []
        for file in FileList:

            file = os.path.abspath(file)
            if self.reportEnabled:
                pm.note ( msg = "processing " + fileSuffix + " file: " + file
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 0
                        )

            restartContents = RestartFileContents   ( file = file
                                                    , reportEnabled = self.reportEnabled
                                                    , methodName = self._methodName
                                                    )
            restartContentsList.append(restartContents)

        outputListName = fileType + "List"
        if renabled:
            outputListFullName = outputListName
            msg = "The processed " + fileType + " files are now stored in the output variable as a\n" \
                + "Python list. For example, to access the contents of the first (or the only) \n" \
                + fileType + " file stored in an output variable named " + outputListFullName + ", try:"
        else:
            outputListFullName = self._objectName + "." + outputListName
            setattr(self, outputListName, restartContentsList)
            msg = "The processed " + fileType + " files are now stored in the newly-created\n" \
                + "component `" + outputListName + "` of the " + self._methodName + " object as a Python list.\n" \
                + "For example, to access the contents of the first (or the only) " + fileType + " file, try:"

        if self.reportEnabled:
            pm.note ( msg   = msg + newline
                            + newline
                            + "    " + outputListFullName + "[0].contents" + newline
                            + newline
                            + "where you will have to replace `pmpd` with your " + self._methodName + " instance name." + newline
                            + "To access the plotting tools, try:" + newline
                            + newline
                            + "    " + outputListFullName + "[0].plot.<PRESS TAB TO SEE THE LIST OF PLOTS>" + newline
                            + newline
                            + "For example," + newline
                            + newline
                            + "    " + outputListFullName + "[0].plot.line()        # to make bivariate line plots." + newline
                            + "    " + outputListFullName + "[0].plot.scatter()     # to make bivariate scatter plots." + newline
                            + "    " + outputListFullName + "[0].plot.lineScatter() # to make bivariate line-scatter plots." + newline
                            + "    " + outputListFullName + "[0].plot.covmat2()     # to make proposal covariance evolution 2D plots." + newline
                            + "    " + outputListFullName + "[0].plot.covmat3()     # to make proposal covariance evolution 3D plots." + newline
                            + "    " + outputListFullName + "[0].plot.cormat2()     # to make proposal correlation evolution 2D plots." + newline
                            + "    " + outputListFullName + "[0].plot.cormat3()     # to make proposal correlation evolution 3D plots." + newline
                            + newline
                            + "For more information and examples on the usage, visit:" + newline
                            + newline
                            + "    " + pm.website.home.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if iswebfile: pm.utils.delFile( file = FileList[0], desc = "the temporarily-downloaded " + fileType + " file", methodName = self._methodName )

        if renabled:
            return restartContentsList
        else:
            return None

    ################################################################################################################################
    #### readReport
    ################################################################################################################################

    def readReport  ( self
                    , file          : tp.Optional[str] = None
                    , renabled      : tp.Optional[bool] = False
                    ) -> tp.List[ReportFileContents] :
        """

        Return a list of the contents of a set of the simulation(s) output
        report files whose names begin the user-provided input file prefix, 
        or as specified by the input simulation specification 
        ``SAMPLER.spec.outputFileName``, where SAMPLER can be 
        an instance of any one of the ParaMonte's sampler 
        classes, such as ``ParaDRAM()``.

        **NOTE**

        This method is to be only used for post-processing of the output
        report file(s) of an already finished simulation. It is NOT meant
        to be called by all processes in parallel mode, although it is
        possible.

            **Parameters**

                file (optional)

                    A string representing the path to the report file with the
                    default value of []. The path only needs to uniquely identify
                    the name of the simulation to which the report file belongs.
                    For example, specifying ``"./mydir/mysim"`` as input will lead to
                    a search for a file that begins with ``"mysim"`` and ends with
                    ``"_report.txt"`` inside the directory ``"./mydir/"``.
                    If there are multiple files with such name, then all of them
                    will be read and returned as a list.
                    If this input argument is not provided by the user, the
                    value of the object's ``spec`` attribute ``outputFileName``
                    will be used instead.

                    **WARNING**

                    At least one of the two mentioned routes must
                    provide the path to the report file. Otherwise,
                    this method will abort the program.

                    Example usage:

                        .. code-block:: python

                            pmpd.readReport("./out/test_run_")

                        or,

                        .. code-block:: python

                            pmpd.spec.outputFileName = "./out/test_run_"
                            pmpd.readReport()

                   Both of the above examples are equivalent.
                   The latter is recommended as it is less confusing.

                renabled (optional)

                    If set to ``False``, the contents of the file(s) will be 
                    stored as a list in a (new) component of the object with a
                    name that ends with the prefix ``List``. Otherwise, ``None`` 
                    will be the return value of the method. If set to ``True``, 
                    the reverse will done. The default value is ``False``.

            **Returns**

                reportList (optional)

                    A Python list of ``ReportFileContents`` objects, each of
                    which corresponds to the contents of a unique report file.
                    Each object may have a dynamic list of the different sections
                    of the output report file
                    Each object may have the following components:

                        file

                            The full absolute path to the report file.

                        contents

                            The contents of the file in its entirely as a string.

                If no output argument is provided, a ``reportList`` property will be
                added to the parent sampler object to which the method ``readReport()``
                belongs.

        """

        fileType = "report"
        fileSuffix = "report"

        if file is None: file = self._setFileToRead(file, fileType, fileSuffix)

        FileList, iswebfile = pm.utils.getFileList(file, fileSuffix, self._methodName, self.reportEnabled)

        reportContentsList = []
        for file in FileList:

            file = os.path.abspath(file)
            if self.reportEnabled:
                pm.note ( msg = "processing " + fileSuffix + " file: " + file
                        , methodName = self._methodName
                        , marginTop = 1
                        , marginBot = 0
                        )

            reportContents = ReportFileContents ( file = file
                                                , reportEnabled = self.reportEnabled
                                                , methodName = self._methodName
                                                )
            reportContentsList.append(reportContents)

        outputListName = fileType + "List"
        if renabled:
            outputListFullName = outputListName
            msg = "The processed " + fileType + " files are now stored in the output variable as a\n" \
                + "Python list. For example, to access the contents of the first (or the only) \n" \
                + fileType + " file stored in an output variable named " + outputListFullName + ", try:"
        else:
            outputListFullName = self._objectName + "." + outputListName
            setattr(self, outputListName, reportContentsList)
            msg = "The processed " + fileType + " files are now stored in the newly-created\n" \
                + "component `" + outputListName + "` of the " + self._methodName + " object as a Python list.\n" \
                + "For example, to access the entire contents of the first (or the only) " + fileType + " file, try:"

        if self.reportEnabled:
            pm.note ( msg   = msg + newline
                            + newline
                            + "    " + outputListFullName + "[0].contents.print()" + newline
                            + newline
                            + "where you will have to replace `pmpd` with your " + self._methodName + " instance name." + newline
                            + "To access the simulation statistics and information, examine the contents of the" + newline
                            + "components of the following structures:" + newline
                            + newline
                            + "    " + outputListFullName + "[0].contents.print()  # to print the contents of the report file." + newline
                            + "    " + outputListFullName + "[0].setup             # to get information about the simulation setup." + newline
                            + "    " + outputListFullName + "[0].stats.time        # to get the timing information of the simulation." + newline
                            + "    " + outputListFullName + "[0].stats.chain       # to get the statistics of the simulation output sample." + newline
                            + "    " + outputListFullName + "[0].stats.numFuncCall # to get information about the number of function calls." + newline
                            + "    " + outputListFullName + "[0].stats.parallelism # to get information about the simulation parallelism." + newline
                            + "    " + outputListFullName + "[0].spec              # to get the simulation specification in the report file." + newline
                            + newline
                            + "For more information and examples on the usage, visit:" + newline
                            + newline
                            + "    " + pm.website.home.url
                    , methodName = self._methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

        if iswebfile: pm.utils.delFile( file = FileList[0], desc = "the temporarily-downloaded " + fileType + " file", methodName = self._methodName )

        if renabled:
            return reportContentsList
        else:
            return None

    ################################################################################################################################
    #### helpme
    ################################################################################################################################

    def helpme  ( self
                , topic : tp.Optional[ str ] = ""
                ):
        """

        Prints help on the input object.

            **Parameters**

                topic

                    A string value that is the name of a component of the current 
                    sample object for which help is needed. For example:  

                    Example usage:

                        .. code-block:: python

                            pm.helpme("helpme")

            **Returns**

                None

        """

        usage   = ("    Usage:" + newline
                + newline
                + "        import paramonte as pm " + newline
                + "        pm.helpme()      # to get help on paramonte module. " + newline
                + "        pm.helpme(topic) # to get help on topic. " + newline
                + newline
                + "    where `topic` in the above can be the name of any " + newline
                + "    component of the current sampler object."
                )

        try:
            doc = eval("self."+topic+".__doc__")
            print(doc + "\n")
        except:
            print(self.__doc__)
            print("\nHere is the information on the parent class:\n")
            print(ParaMonteSampler.__doc__)

        if topic.lower()=="helpme": pm.note( msg = usage, methodName = "helpme()", marginTop = 0, marginBot = 1)

        return None

    ################################################################################################################################



#!/usr/bin/python
# Author:  Amir Shahmoradi
# Contact: shahmoradi@utexas.edu
####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################
"""

This is the Python interface to ParaMonte: Plain Powerful Parallel Monte Carlo library.

What is ParaMonte?
==================

ParaMonte is a serial / parallel library of Monte Carlo routines for sampling 
mathematical objective functions of arbitrary-dimensions, in particular, 
the posterior distributions of Bayesian models in data science, 
Machine Learning, and scientific inference, with the design 
goal of unifying the

    **automation** of Monte Carlo simulations,  

    **user-friendliness** of the library,  

    **accessibility** from multiple programming environments,  

    **high-performance** at runtime, and,  

    **scalability** across many parallel processors.  

For more information on the installation, usage, and examples, visit:

    https://www.cdslab.org/paramonte

For the API documentation, visit:

    https://www.cdslab.org/paramonte/notes/api/python

ParaMonte samplers
================== 

The routines currently supported by the ParaMonte Python library include:

    **ParaDRAM**

        Parallel Delayed-Rejection Adaptive Metropolis-Hastings Markov 
        Chain Monte Carlo Sampler. For a quick start, example scripts, 
        and instructions on how to use he ParaDRAM sampler, type the 
        following commands enclosed between the comment lines 
        in your Python session,

        .. code-block:: python
            :linenos:

            ##################################
            import paramonte as pm
            pm.helpme("paradram") # the input value is case-insensitive
            ##################################

        or,

        .. code-block:: python
            :linenos:

            ##################################
            import paramonte as pm
            help(pm.ParaDRAM) # get help on ParaDRAM sampler class
            ##################################

Naming conventions
==================

+   The camelCase naming style is used throughout the entire ParaMonte 
    library, across all programming languages. The ParaMonte library is 
    a multi-language cross-platform library. To increase the consistently 
    and similarities of all implementations, a single naming convention 
    had to be used for all different languages.

+   All simulation specifications start with a lowercase letter, including 
    scalar/vector/matrix int, float, string, or boolean variables.

+   The name of any variable that represents a vector of values is normally 
    suffixed with ``Vec``, for example: ``startPointVec``, ...

+   The name of any variable that represents a matrix of values is normally 
    suffixed with ``Mat``, for example: ``proposalStartCorMat``, ...

+   The name of any variable that represents a list of varying-size values 
    is normally suffixed with ``List``, like: ``variableNameList``, ...

+   All static functions or methods of classes begin with a lowercase verb.

+   Significant attempt has been made to end all boolean variables with a 
    passive verb, such that the full variable name virtually forms a 
    proposition, that is, an English-language statement that should 
    be either ``True`` or ``False``, set by the user. 

Tips
====

+   When running the ParaMonte samplers, in particular on multiple cores 
    in parallel, it would be best to close any such aggressive software or
    applications as **Dropbox**, **ZoneAlarm**, ... that can interfere with 
    the ParaMonte simulation output files, potentially causing the sampler to
    crash before the successful completion of the simulation. 
    These situations should however happen only scarcely. 

+   On Windows systems, when restarting an old interrupted ParaMonte simulation, 
    ensure your Python session is also restarted before the simulation restart. 
    This may be needed as Windows sometimes locks access to some or all of the 
    simulation output files.

+   To unset an already-set input simulation specification, simply set the 
    simulation attribute to None or re-instantiate the sampler object. 

-------------------------------------------------------------------------------
"""

import os as _os
import sys as _sys
import typing as _tp

_sys.path.append(_os.path.dirname(__file__))

import _paramonte as _pm

#### objects exposed to the user

from _pmreqs import verify, build, checkForUpdate, verifyDependencyVersion, getDependencyVersion
from _paradram import ParaDRAM
from _paramonte import version, website, cite


__authors__ = "The Computational Data Science Lab @ The University of Texas"
__credits__ = "The Peter O'Donnell Fellowship"
__version__ = version.interface.dump()
"""ParaMonte Python interface version, defined at runtime."""

verify(reset=False)

####################################################################################################################################

# def getVersion():
#     import warnings
#     msg = "getVersion() version will be removed in next release. Use paramonte.version object instead."
#     warnings.warn( msg, DeprecationWarning, stacklevel=2)
#     return __version__

####################################################################################################################################

def helpme  ( topic : _tp.Optional[ str ] = None
            ):
    """

    Prints help on the input object.

        **Parameters**

            topic

                A string value that is the name of an object in ``paramonte``
                module for which help is needed. To see the list of possible
                objects. try: ``pm.helpme("helpme")``

        **Returns**

            None

    """

    topics =    { "paradram": ParaDRAM
                , "version" : version
                , "verify"  : verify
                , "helpme"  : helpme
                , "build"   : build
                }

    usage   = "    Usage:\n\n" \
            + "        import paramonte as pm\n" \
            + "        pm.helpme()      # to get help on paramonte module.\n" \
            + "        pm.helpme(topic) # to get help on topic.\n\n" \
            + "    where `topic` in the above can be one of the following string values:\n\n" \
            + "        " + str(list(topics.keys()))
    
    if topic is None:
        print(__doc__)
    elif isinstance(topic,str) and (topic.lower() in topics.keys()):
        print(topics[topic.lower()].__doc__ + "\n")
        if topic.lower()=="helpme": _pm.note( msg = usage, methodName = "helpme()", marginTop = 0, marginBot = 1)
    else:
        try:
            topic = "(" + str(topic) + ") "
        except:
            topic = ""
        _pm.warn( msg   = "The requested object " + topic + "does not exist in paramonte module.\n\n"
                        + usage
                , methodName = "helpme()"
                , marginTop = 1
                , marginBot = 1
                )

    return None

####################################################################################################################################

__all__ =   [ "__authors__"
            , "__credits__"
            , "__version__"
            , "ParaDRAM"
            , "version"
            , "verify"
            , "helpme"
            , "build"
            ]



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import _paramonte as pm

####################################################################################################################################
#### OutputFileContents
####################################################################################################################################

class OutputFileContents:
    """

    This is the **OutputFileContents** base class for the ParaMonte 
    sampler output file contents classes. **This class is NOT meant to be 
    directly accessed or called by the user of the ParaMonte library.**
    However, its children, such the ParaDRAM sampler class will be 
    indirectly accessible to the public. 

        **Parameters**

            file

                The full path to the file.

            methodName

                A string representing the name of the ParaMonte sampler used
                to call the constructor of the ``OutputFileContents`` class.

            reportEnabled

                A logical input parameter indicating whether the ParaMonte
                automatic guidelines to the standard output should be provided 
                or not. The default value is ``True``.

        **Methods**

            See below for information on the methods.  

        **Returns**

            Object of class OutputFileContents

    """

    def __init__( self
                , file          : str
                , methodName    : str
                , reportEnabled : bool
                ):

        self.file = file
        self._methodName = methodName
        self._reportEnabled = reportEnabled
        self._progress = pm.utils.Progress  ( msg = "reading the file contents... "
                                            , methodName = methodName
                                            , reportEnabled = reportEnabled
                                            , end = ""
                                            )

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import _paramonte as pm
from _OutputFileContents import OutputFileContents

Struct = pm.Struct
newline = pm.newline
creturn = pm.creturn

class StructWithPrint:
    def __init__(self): self.value = None
    def print(self): print(self.value)

####################################################################################################################################
#### ReportFileContents
####################################################################################################################################

class ReportFileContents(OutputFileContents):
    """

    This is the **ReportFileContents** class for generating instances
    of the ParaMonte output report file contents. This class is NOT meant 
    to be directly accessed by the ParaMonte library users. It is internally
    used by the ParaMonte library to parse the contents of the output Report
    files generated by the ParaMonte sampler routines. For example, the ParaDRAM
    sampler class makes calls to this class via its ``readReport()`` method to
    return a list of objects of class ``ReportFileContents``.

        **Parameters**

            file

                The full path to the file containing the report.

            methodName

                A string representing the name of the ParaMonte sampler used
                to call the constructor of the ``ReportFileContents`` class.

            reportEnabled

                A logical input parameter indicating whether the ParaMonte
                automatic guidelines to the standard output should be provided 
                or not. The default value is ``True``.

        **Attributes**

            A dynamic set of attributes that are directly parsed from the file.

        **Returns**

            ReportFileContents

                An object of class ``ReportFileContents``.

    ----------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , file
                , methodName
                , reportEnabled
                ):

        super().__init__(file, methodName, reportEnabled)

        # set the hidden variables

        self._lineListLen = []
        self._indentLen = 8  # indent length of the records
        self._dsym = "****"  # decoration symbol
        self._lineCounter = 0
        self._prefix = Struct()
        self._dsymLen = len(self._dsym)
        self._prefix.note = self._methodName + " - NOTE: "
        self._prefix.warn = self._methodName + " - WARNING: "
        self.contents = StructWithPrint()

        with open(self.file,"r") as fid: self.contents.value = fid.read().replace(creturn,"").replace(newline, " " + newline)
        self._lineList = self.contents.value.splitlines()
        self._lineListLen = len(self._lineList)

        self._progress.note( msg = "parsing the report file contents...", end = newline, pre = True )

        self._lineCounter  = -1

        ############################################################################################################################
        #### define the components
        ############################################################################################################################

        self.setup = Struct()
        self.setup.library = Struct()
        self.setup.library.banner = StructWithPrint()
        self.setup.library.interface = StructWithPrint()
        self.setup.platform = StructWithPrint()
        self.setup.io = StructWithPrint()
        self.spec = StructWithPrint()

        ############################################################################################################################
        #### read the banner
        ############################################################################################################################

        lineStartFound = False
        lineStart = 0
        while True:
            self._lineCounter += 1
            if self._lineCounter>=self._lineListLen: break
            if lineStartFound:
                if self._dsym not in self._lineList[self._lineCounter]:
                    self._lineCounter -= 1
                    break
            else:
                if self._dsym in self._lineList[self._lineCounter]:
                    lineStart = self._lineCounter
                    lineStartFound = True

        self.setup.library.banner.description = "This is the ParaMonte banner appearing at the beginning of each file."
        if lineStartFound:
            self.setup.library.banner.value = self._concat(lineStart,self._lineCounter)
        else:
            self._reportParseFailure("ParaMonte banner")

        ############################################################################################################################
        #### read the ParaMonte library interface specifications
        ############################################################################################################################

        self.setup.library.interface.description = "This is the interface to the ParaMonte library that has been used to generate this report file."
        self.setup.library.interface.value = self._parseSection("ParaMonte library interface specifications")

        ############################################################################################################################
        #### read the Runtime platform specifications
        ############################################################################################################################

        self.setup.platform.description = "Information about the runtime platform specifications"
        self.setup.platform.value = self._parseSection("Runtime platform specifications")

        ############################################################################################################################
        #### read the simulation environment
        ############################################################################################################################

        self.setup.io.description = "Information about the simulation input and output files."
        self.setup.io.value = self._parseSection("simulation environment")

        ############################################################################################################################
        #### read the simulation specifications
        ############################################################################################################################

        self.spec.description = "Information about all input simulation specifications."
        self.spec.value = self._parseSection("simulation specifications").replace(self._prefix.note,"").replace(self._prefix.warn,"")

        ############################################################################################################################
        #### statistics: this must be always the last item to parse
        ############################################################################################################################

        self._parseStats()

        ############################################################################################################################

        self._progress.note()

    ################################################################################################################################
    #### _reportParseFailure
    ################################################################################################################################

    def _reportParseFailure(self,topic):
        topic = topic.replace(newline," ").replace(creturn," ").strip()
        pm.warn ( msg   = "Failed to parse the record \"" + topic + "\". "
                        + "The structure of the report file appears to have been compromised. skipping... "
                        , methodName = self._methodName
                        , marginTop = 0
                        , marginBot = 0
                        )

    ################################################################################################################################
    #### _reportMissingValue
    ################################################################################################################################

    def _reportMissingValue(self,topic):
        topic = topic.replace(newline," ").replace(creturn," ").strip()
        pm.warn ( msg   = "Failed to parse the value corresponding to the record \"" + topic + "\". "
                        + "The structure of the report file appears to have been compromised. skipping... "
                        , methodName = self._methodName
                        , marginTop = 0
                        , marginBot = 0
                        )

    ################################################################################################################################
    #### _isSectionHeader
    ################################################################################################################################

    def _isSectionHeader(self,record): return True if len(record)>self._dsymLen and self._dsym in record[0:self._dsymLen] else False

    ################################################################################################################################
    #### _stopBeforeNextSectionHeader
    ################################################################################################################################

    def _stopBeforeNextSectionHeader(self):
        # NOTE: if it is already within a section header, it will skip it and find the next section header.
        record = self._lineList[self._lineCounter]
        isCurrentSectionHeader = len(record)>self._dsymLen and self._dsym in record[0:self._dsymLen+1]
        while True:
            if self._lineCounter==self._lineListLen: break
            record = self._lineList[self._lineCounter]
            if len(record)>self._dsymLen and self._dsym in record[0:self._dsymLen]:
                if isCurrentSectionHeader:
                    self._lineCounter += 1
                    continue
                else:
                    self._lineCounter -= 1
                    break
            else:
                isCurrentSectionHeader = False
                self._lineCounter += 1
                continue

    ################################################################################################################################
    #### _skipCurrentSectionHeader
    ################################################################################################################################

    def _skipCurrentSectionHeader(self):
        while True:
            record = self._lineList[self._lineCounter]
            if len(record)>self._dsymLen and self._dsym in record[0:self._dsymLen]:
                self._lineCounter += 1
                if self._lineCounter==self._lineListLen: break
                continue
            else:
                break

    ################################################################################################################################
    #### _parseSection
    ################################################################################################################################

    def _parseSection(self,topic):
        lineCounterLastSuccess = self._lineCounter
        section = []
        topicFound = False
        while True:
            self._lineCounter += 1
            if self._lineCounter>=self._lineListLen: break
            record = self._lineList[self._lineCounter]
            if topic in record:
                topicFound = True
                break
            else:
                continue

        if topicFound:
            self._skipCurrentSectionHeader()
            lineStart = self._lineCounter
            self._stopBeforeNextSectionHeader()
            section = self._concat(lineStart, self._lineCounter)
        else:
            self._reportParseFailure(topic)
            self._lineCounter = lineCounterLastSuccess

        return section

    ################################################################################################################################
    #### _concat
    ################################################################################################################################

    def _concat(self,lineStart,lineCounter): return newline.join(self._lineList[lineStart:lineCounter+1])

    ################################################################################################################################
    #### _parseStats
    ################################################################################################################################

    def _parseStats(self):

        self.stats = Struct()
        if self._methodName=="ParaDRAM" or self._methodName=="ParaDISE" or self._methodName=="MatDRAM":

            while True:

                self._lineCounter += 1
                if self._lineCounter>=self._lineListLen: break
                item = self._lineList[self._lineCounter]

                if self._isstats(item):

                    ################################################################################################################
                    #### parse the value
                    ################################################################################################################


                    valueIsNumeric = True
                    valueFound = False
                    descFound = False
                    value = ""
                    desc = ""

                    ################################################################################################################

                    while True:

                        self._lineCounter += 1
                        if self._lineCounter>=self._lineListLen: break
                        record = self._lineList[self._lineCounter].replace(newline," ").replace(creturn," ")

                        #### check the record is not another item or desc is not found before value.

                        if self._isstats(record): # and not (descFound and valueFound)
                            self._lineCounter -= 1
                            break

                        #### parse the value/description

                        if len(record)>self._indentLen:
                            recordIsDesc = self._isdesc(record)
                            if recordIsDesc:
                                if not valueFound:
                                    self._reportMissingValue(item)
                                    break
                                descFound = True
                                desc += " " + self._lineList[self._lineCounter].replace(self._prefix.note," ").replace(self._prefix.warn," ").strip() # remove prefix, trim, append.
                            elif not descFound:
                                valueFound = True
                                value += record
                                try:
                                    _ = [ np.double(element) for element in record.split() ] # numericValue
                                except:
                                    valueIsNumeric = False


                    ################################################################################################################

                    if valueFound and descFound:
                        componentList = item.split(".")
                        currentComponentName = "self"
                        for component in componentList:
                            propExists = eval( "hasattr(" + currentComponentName + ",'" + component + "')" )
                            if not propExists:
                                exec(currentComponentName+"."+component+"=Struct()")
                            currentComponentName += "." + component
                        if valueIsNumeric and "infinity" not in value.lower():
                            value = value.split()
                            value = value[0] if len(value)==1 else "[" + ",".join(value) + "]" # make vector of values
                        else:
                            value = "r'" + value + "'"
                            exec(currentComponentName + "=StructWithPrint()")
                        exec( currentComponentName + ".value=" + value )
                        exec( currentComponentName + ".description=r'" + desc.strip() + "'" )

    ################################################################################################################################
    #### _isstats
    ################################################################################################################################

    def _isstats(self,record): return len(record)>5 and record[0:6]=="stats."

    ################################################################################################################################
    #### _isstats
    ################################################################################################################################

    def _isdesc(self,record): return self._prefix.note in record or self._prefix.warn in record



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import typing as tp
import weakref as wref
import _paramonte as pm
import _pmutils as pmutils
from paramonte.vis.Target import Target
from paramonte.vis._BasePlot import BasePlot

Struct = pmutils.Struct
newline = pmutils.newline

####################################################################################################################################
#### LineScatterPlot class
####################################################################################################################################

class LineScatterPlot(BasePlot):
    """

    This is the LineScatterPlot class for generating instances
    of line or scatter plots or the combination of the two in
    two or three dimensions based on the visualization tools
    of the ``matplotlib`` and ``seaborn`` Python libraries.

        **Usage**

            First generate an object of this class by optionally
            passing the following parameters described below. Then call
            the ``make()`` method. The generated object is also callable 
            with the same input parameters as the object's constructor.

        **Parameters**

            plotType

                A string indicating the name of the plot to be constructed.

            dataFrame (optional)

                A pandas dataFrame whose data will be plotted.

            methodName (optional)

                The name of the ParaMonte sample requesting the BasePlot.

            reportEnabled (optional)

                A boolean whose value indicates whether guidelines should be 
                printed in the standard output.

            resetPlot (optional)

                A function that resets the properties of the plot as desired 
                from outside. If provided, a pointer to this function will be
                saved for future internal usage.

        **Attributes**

            xcolumns

                An attribute that determines the columns of dataFrame 
                to be visualized as the X-axis. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``xcolumns = [0,1,4,3]``
                    2.  ``xcolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``xcolumns = range(17,7,-2)``

                The default behavior includes all columns of the dataFrame.

            ycolumns

                An attribute that determines the columns of dataFrame 
                to be visualized as the Y-axis. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``ycolumns = [0,1,4,3]``
                    2.  ``ycolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``ycolumns = range(17,7,-2)``

                The default behavior includes all columns of the dataFrame.

            zcolumns (exists only in 3D plot objects)

                An attribute that determines the columns of dataFrame 
                to be visualized as the Z-axis. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``zcolumns = [0,1,4,3]``
                    2.  ``zcolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``zcolumns = range(17,7,-2)``

                The default behavior includes all columns of the dataFrame.

            ccolumns

                An attribute that determines the columns of dataFrame 
                to be used for color mapping. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``ccolumns = [0,1,4,3]``
                    2.  ``ccolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``ccolumns = range(17,7,-2)``

                If ``ccolumns`` is set to ``None``, then no color-mapping 
                will be made. If it is set to an empty list ``[]``, then 
                the values from the ``rows`` attribute will be used for 
                color-mapping.

            rows

                An attribute that determines the rows of dataFrame to be 
                visualized. It can be either:

                    1.  A ``range(start,stop,step)``, or, 
                    2.  A list of row indices in dataFrame.index.

                Examples:

                    1.  ``rows = range(17,7,-2)``
                    2.  ``rows = [i for i in range(7,17)]``

                The default behavior includes all rows of the dataFrame.

            plot (exists only for line or lineScatter plots in 2D and 3D)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``plot()``
                        function of the matplotlib library should be made 
                        or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``plot()`` function.

                        Example usage:

                            .. code-block:: python

                                plot.enabled = True
                                plot.kws.linewidth = 1

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            scatter (exists only for scatter / lineScatter plots in 2D and 3D)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the 
                        ``scatter()`` function of the matplotlib library 
                        should be made or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``scatter()`` function.

                        Example usage:

                            .. code-block:: python

                                scatter.enabled = True
                                scatter.kws.s = 2

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            lineCollection (exists only for 2D / 3D line / lineScatter plots)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the 
                        ``LineCollection()`` class of the matplotlib 
                        library should be made or not. This will result 
                        in line plots that are color-mapped.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``LineCollection()`` class 
                        of matplotlib library.

                        Example usage:

                            .. code-block:: python

                                lineCollection.enabled = True
                                lineCollection.kws.linewidth = 1

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            set

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``set()``
                        function of the seaborn library should be made or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set()`` function.

                        Example usage:

                            .. code-block:: python

                                set.kws.style = "darkgrid"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes (available only in 1D and 2D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``gca()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes3d (available only in 3D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``Axes3D()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes3d.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            figure

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``figure()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``figure()`` function.

                        Example usage:

                            .. code-block:: python

                                figure.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            colorbar (exists only for plots that require colorbar)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``colorbar()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``colorbar()`` function of 
                        the matplotlib library.

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A colorbar will be added to a plot only if a color-mappings 
                is requested in the plot.

            legend (may not exist for some types of plots)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``legend()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``legend()`` function.

                        Example usage:

                        .. code-block:: python

                            legend.kws.labels = ["Variable1", "Variable2"]

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A legend will be added to a plot only if no color-mappings are
                requested in the plot.

            currentFig

                A structure whose attributes are the outputs of various plotting 
                tools used to make the current figure. These include the handle 
                to the current figure, the handle to the current axes in the plot, 
                the handle to the colorbar (if any exists), and other Python 
                plotting tools used to make to generate the figure.

            target (available only in 1D and 2D plot objects)

                A callable object of the ParaMonte library's ``Target`` class 
                which can be used to add target point or lines to the current 
                active plot.

        **Returns**

                An object of class ``LineScatterPlot``.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , plotType      # : str
                , dataFrame     = None # : tp.Optional[ pd.DataFrame ] = None
                , methodName    = "ParaMonte" # : tp.Optional[ str ] = "ParaMonte"
                , reportEnabled = True # : tp.Optional[ bool ] = True
                , resetPlot     = None
                ):

        super().__init__( plotType = plotType
                        , dataFrame = dataFrame
                        , methodName = methodName
                        , reportEnabled = reportEnabled
                        , resetPlot = resetPlot
                        )

        self._reset()
        if resetPlot is None: self._resetPlot = self._reset
        self._progress.note()

    ################################################################################################################################
    #### _reset
    ################################################################################################################################

    def _reset(self):

        super()._reset()
        self._indexOffset       = 1
        self.xcolumns           = None
        self.ycolumns           = None
        self.ccolumns           = [] # indicates the use of default column for colormap. To turn off

        self._xlabel = None
        self._ylabel = None
        self._zlabel = None

        self._xlimit = None
        self._ylimit = None
        self._zlimit = None

        self._xscale = None
        self._yscale = None
        self._zscale = None

        if not self._type.is3d: self.target = Target()

        if self._type.isLine:
            self.plot = Struct()
            self.plot.enabled = False
            self.plot.kws = Struct()
            self.lineCollection = Struct()
            self.lineCollection.kws = Struct()
            self.lineCollection.enabled = True

        if self._type.isScatter:
            self.scatter = Struct()
            self.scatter.enabled = True
            self.scatter.kws = Struct()

        if self._type.is3d:
            self.zcolumns = []

        self.colorbar = Struct()
        self.colorbar.kws = Struct()
        if (self._type.isScatter and self.ccolumns is not None) or (self._type.isLine and self.lineCollection.enabled):
            self.colorbar.enabled = True
        else:
            self.colorbar.enabled = False

        self._isdryrun = True
        self.make()
        self._isdryrun = False

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``make()`` method of the current 
        instance of the class.

            **Parameters**

                Any arguments that can be passed to the 
                ``make()`` method of the plot object.

            **Returns**

                Any return value from the ``make()`` 
                method of the plot object.

        """

        return self.make(reself, **kwargs)

    ################################################################################################################################
    #### make
    ################################################################################################################################

    def make( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Generate a line/scatter plot from the 
        selected columns of the object's dataframe.

            **Parameters**

                reself

                    A logical variable. If ``True``, an instance of 
                    the object will be returned to the calling routine 
                    upon exit. The default value is ``False``.

            **Returns**

                The object self if ``reself = True`` otherwise, ``None``.
                However, this method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key in kwargs.keys():
            if hasattr(self,key):
                setattr(self, key, kwargs[key])
            elif key=="dataFrame":
                setattr( self, "_dfref", wref.ref(kwargs[key]) )
            else:
                raise Exception ( "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        # set what to plot

        cEnabled = self.ccolumns is not None

        from collections.abc import Iterable
        if self.ccolumns is not None and not isinstance(self.ccolumns, Iterable): self.ccolumns = [self.ccolumns]

        # if no colormap, then 

        if self._type.isLine and not cEnabled: self.plot.enabled = True

        ############################################################################################################################
        #### scatter plot properties
        ############################################################################################################################

        if self._type.isScatter:
            if isinstance(self.scatter.kws,Struct):

                if "s" not in vars(self.scatter.kws).keys(): self.scatter.kws.s = 2
                if "c" not in vars(self.scatter.kws).keys(): self.scatter.kws.c = None
                if "cmap" not in vars(self.scatter.kws).keys() or self.scatter.kws.cmap is None: self.scatter.kws.cmap = "autumn"
                if "alpha" not in vars(self.scatter.kws).keys(): self.scatter.kws.alpha = 1
                if "edgecolors" not in vars(self.scatter.kws).keys(): self.scatter.kws.edgecolors = None
                if "zorder" not in vars(self.scatter.kws).keys(): self.scatter.kws.zorder = 2

                if not cEnabled: self.scatter.kws.cmap = None

            else:

                raise Exception ( "The scatter.kws component of the current LineScatterPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the scatter() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### line plot properties
        ############################################################################################################################

        if self._type.isLine:

            if isinstance(self.plot.kws,Struct):

                if "linewidth" in vars(self.plot.kws).keys():
                    if self.plot.kws.linewidth==0: self.plot.kws.linewidth = 1
                else:
                    self.plot.kws.linewidth = 1

                if "zorder" not in vars(self.plot.kws).keys(): self.plot.kws.zorder = 1

            else:

                raise Exception ( "The plot.kws component of the current LineScatterPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the plot() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

            if isinstance(self.lineCollection.kws, Struct):

                if "cmap" not in vars(self.lineCollection.kws).keys() or self.lineCollection.kws.cmap is None: self.lineCollection.kws.cmap = "autumn"
                if "alpha" not in vars(self.lineCollection.kws).keys(): self.lineCollection.kws.alpha = 1
                if "linewidth" not in vars(self.lineCollection.kws).keys(): self.lineCollection.kws.linewidth = 1

            else:

                objectType = "LineCollection"
                if self._type.is3d: objectType = "Line3DCollection"
                raise Exception ( "The lineCollection.kws component of the current LineScatterPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the " + objectType + "() class of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### legend properties
        ############################################################################################################################

        if self.legend.enabled:
            if not isinstance(self.legend.kws,Struct):
                raise Exception ( "The legend.kws component of the current LineScatterPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the legend() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### figure properties
        ############################################################################################################################

        if self.figure.enabled:
            if isinstance(self.figure.kws, Struct):
                if "dpi" not in vars(self.figure.kws).keys(): self.figure.kws.dpi = 150
                if "facecolor" not in vars(self.figure.kws).keys(): self.figure.kws.facecolor = "w"
                if "edgecolor" not in vars(self.figure.kws).keys(): self.figure.kws.edgecolor = "w"
            else:
                raise Exception ( "The figure.kws component of the current LineScatterPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the figure() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        ############################################################################################################################
        if self._isdryrun: return
        ############################################################################################################################
        ############################################################################################################################

        from matplotlib import pyplot as plt
        from matplotlib.collections import LineCollection
        if self._type.is3d: from mpl_toolkits.mplot3d.art3d import Line3DCollection

        ############################################################################################################################
        #### generate figure and axes if needed
        ############################################################################################################################

        self._constructBasePlot()

        ############################################################################################################################
        #### check data type
        ############################################################################################################################

        self._checkDataType()

        ############################################################################################################################
        #### check rows presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.rows is None: self.rows = range(len(self._dfref().index))

        ############################################################################################################################
        #### check columns presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.xcolumns is None:
            lgxicol = 0
            xcolindex = []
            xcolnames = ["Count"]
            if self._type.isScatter : self.scatter._xvalues = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
            if self._type.isLine    : self.plot._xvalues    = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
        else:
            xcolnames, xcolindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.xcolumns)

        if self.ycolumns is None:
            lgyicol = 0
            ycolindex = []
            ycolnames = ["Count"]
            if self._type.isScatter : self.scatter._yvalues = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
            if self._type.isLine    : self.plot._yvalues    = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
        else:
            ycolnames, ycolindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.ycolumns)

        if self._type.is3d:
            if self.zcolumns is None:
                lgzicol = 0
                zcolindex = []
                zcolnames = ["Count"]
                if self._type.isScatter : self.scatter._zvalues = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
                if self._type.isLine    : self.plot._zvalues    = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
            else:
                zcolnames, zcolindex = pm.dfutils.getColNamesIndex(self._dfref().columns, self.zcolumns)

        ############################################################################################################################
        #### set colormap data
        ############################################################################################################################

        if cEnabled:
            if len(self.ccolumns)==0:
                ccolindex = []
                ccolnames = ["Count"]
                if self._type.isScatter : self.scatter.kws.c    = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
                if self._type.isLine    : cdata                 = np.array( self._dfref().index[self.rows] + self._indexOffset ).flatten()
            else:
                ccolnames, ccolindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.ccolumns)
        else:
            ccolindex = []
            ccolnames = []
            #self.scatter.kws.c = None

        ############################################################################################################################
        #### check the consistency of the lengths
        ############################################################################################################################

        xcolindexlen = len(xcolindex)
        ycolindexlen = len(ycolindex)
        ccolindexlen = len(ccolindex)

        maxLenColumns = np.max  (   [ xcolindexlen
                                    , ycolindexlen
                                    , ccolindexlen
                                    ]
                                )

        if xcolindexlen!=maxLenColumns and xcolindexlen>1: raise Exception("length of xcolumns must be either 1 or equal to the lengths of ycolumns or ccolumns.")
        if ycolindexlen!=maxLenColumns and ycolindexlen>1: raise Exception("length of ycolumns must be either 1 or equal to the lengths of xcolumns or ccolumns.")
        if ccolindexlen!=maxLenColumns and ccolindexlen>1: raise Exception("length of ccolumns must be either 1 or equal to the lengths of xcolumns or ycolumns.")

        if self._type.is3d:
            zcolindexlen = len(zcolindex)
            if zcolindexlen!=maxLenColumns and zcolindexlen>1: raise Exception("length of zcolumns must be either 1 or equal to the lengths of xcolumns or ycolumns.")

        ############################################################################################################################
        #### assign data in case of single column assignments
        ############################################################################################################################

        if xcolindexlen==1:
            lgxicol = 0
            if self._type.isScatter     : self.scatter._xvalues = self._dfref().iloc[self.rows,xcolindex].values.flatten()
            if self._type.isLine        : self.plot._xvalues    = self._dfref().iloc[self.rows,xcolindex].values.flatten()

        if ycolindexlen==1:
            lgyicol = 0
            if self._type.isScatter     : self.scatter._yvalues = self._dfref().iloc[self.rows,ycolindex].values.flatten()
            if self._type.isLine        : self.plot._yvalues    = self._dfref().iloc[self.rows,ycolindex].values.flatten()

        if self._type.is3d:
            if zcolindexlen==1:
                lgzicol = 0
                if self._type.isScatter : self.scatter._zvalues = self._dfref().iloc[self.rows,zcolindex].values.flatten()
                if self._type.isLine    : self.plot._zvalues    = self._dfref().iloc[self.rows,zcolindex].values.flatten()

        if cEnabled:
            if ccolindexlen==1:
                if self._type.isScatter : self.scatter.kws.c    = self._dfref().iloc[self.rows,ccolindex].values.flatten()
                if self._type.isLine    : cdata                 = self._dfref().iloc[self.rows,ccolindex].values.flatten()

        ############################################################################################################################
        #### add line/scatter plot
        ############################################################################################################################

        if self.legend.enabled: self.legend._labels = []
        if self._type.isScatter and self.scatter.enabled: self.currentFig.scatterList = []
        if self._type.isLine:
            if self.plot.enabled: self.currentFig.plotList = []
            if cEnabled and self.lineCollection.enabled:
                self.currentFig.lineCollectionList = []
                self.currentFig.plotList = []

        for i in range(maxLenColumns):

            if xcolindexlen>1:
                lgxicol = i
                if self._type.isScatter     : self.scatter._xvalues = self._dfref().iloc[self.rows,xcolindex[i]].values.flatten()
                if self._type.isLine        : self.plot._xvalues    = self._dfref().iloc[self.rows,xcolindex[i]].values.flatten()

            if ycolindexlen>1:
                lgyicol = i
                if self._type.isScatter     : self.scatter._yvalues = self._dfref().iloc[self.rows,ycolindex[i]].values.flatten()
                if self._type.isLine        : self.plot._yvalues    = self._dfref().iloc[self.rows,ycolindex[i]].values.flatten()

            if cEnabled:
                if ccolindexlen>1:
                    if self._type.isScatter : self.scatter.kws.c    = self._dfref().iloc[self.rows,ccolindex[i]].values.flatten()
                    if self._type.isLine    : cdata                 = self._dfref().iloc[self.rows,ccolindex[i]].values.flatten()

            if self.legend.enabled:
                if xcolindexlen<2 and ycolindexlen>1:
                    self.legend._labels.append(ycolnames[lgyicol])
                elif xcolindexlen>1 and ycolindexlen<2:
                    self.legend._labels.append(xcolnames[lgxicol])
                else:
                    self.legend._labels.append( xcolnames[lgxicol] + "-" + ycolnames[lgyicol] )

            if self._type.is3d:
                if zcolindexlen>1:
                    lgzicol = i
                    if self._type.isScatter : self.scatter._zvalues = self._dfref().iloc[self.rows,zcolindex[i]].values.flatten()
                    if self._type.isLine    : self.plot._zvalues    = self._dfref().iloc[self.rows,zcolindex[i]].values.flatten()
                if self.legend.enabled:
                    if zcolindexlen>1: self.legend._labels[-1] += "-" + zcolnames[lgzicol]

            ########################################################################################################################
            #### add scatter plot
            ########################################################################################################################

            if self._type.isScatter and self.scatter.enabled:

                if self._type.is3d:

                    self.currentFig.scatterList.append( self.currentFig.axes.scatter( self.scatter._xvalues
                                                                                    , self.scatter._yvalues
                                                                                    , self.scatter._zvalues
                                                                                    , **vars(self.scatter.kws)
                                                                                    ) )

                else:

                    self.currentFig.scatterList.append( self.currentFig.axes.scatter( self.scatter._xvalues
                                                                                    , self.scatter._yvalues
                                                                                    , **vars(self.scatter.kws)
                                                                                    ) )

            ########################################################################################################################
            #### add line plot
            ########################################################################################################################

            if self._type.isLine:

                if self.plot.enabled:

                    if self._type.is3d:

                        self.currentFig.plotList.append ( self.currentFig.axes.plot ( self.plot._xvalues
                                                                                    , self.plot._yvalues
                                                                                    , self.plot._zvalues
                                                                                    , **vars(self.plot.kws)
                                                                                    ) )

                    else:

                        self.currentFig.plotList.append ( self.currentFig.axes.plot ( self.plot._xvalues
                                                                                    , self.plot._yvalues
                                                                                    , **vars(self.plot.kws)
                                                                                    ) )

                if cEnabled and self.lineCollection.enabled:

                    self.lineCollection.kws.norm = norm = plt.Normalize(cdata.min(), cdata.max())

                    if self._type.is3d:

                        # properly and automatically set the axes limits via plot()

                        self.currentFig.plotList.append ( self.currentFig.axes.plot ( self.plot._xvalues
                                                                                    , self.plot._yvalues
                                                                                    , self.plot._zvalues
                                                                                    , linewidth = 0
                                                                                    ) )

                        # now add the lineCollection

                        points = np.array([self.plot._xvalues, self.plot._yvalues, self.plot._zvalues]).T.reshape(-1, 1, 3)
                        segments = np.concatenate([points[:-1], points[1:]], axis=1)
                        lineCollection = Line3DCollection( segments, **vars(self.lineCollection.kws) )

                    else:

                        # properly and automatically set the axes limits via plot()

                        self.currentFig.plotList.append ( self.currentFig.axes.plot ( self.plot._xvalues
                                                                                    , self.plot._yvalues
                                                                                    , linewidth = 0
                                                                                    ) )

                        # now add the lineCollection

                        points = np.array([self.plot._xvalues, self.plot._yvalues]).T.reshape(-1, 1, 2)
                        segments = np.concatenate([points[:-1], points[1:]], axis=1)
                        lineCollection = LineCollection( segments, **vars(self.lineCollection.kws) )

                    lineCollection.set_array(cdata)
                    #lineCollection.set_linewidth(0.5)
                    #lineCollection.set_solid_capstyle("round")
                    self.currentFig.lineCollectionList.append( self.currentFig.axes.add_collection(lineCollection) )

        ############################################################################################################################
        #### add colorbar
        ############################################################################################################################

        cbarEnabled = cEnabled and self.colorbar.enabled and (ccolindexlen<2) # and (not hasattr(self.currentFig,"colorbar"))
        if cbarEnabled:

            self.colorbar.kws.mappable = None
            if self._type.isLine and self.lineCollection.enabled:

                self.colorbar.kws.mappable = self.currentFig.lineCollectionList[0]

            elif self._type.isScatter and self.scatter.enabled:

                self.colorbar.kws.mappable = self.currentFig.scatterList[0]

            if self.colorbar.kws.mappable is not None:
                self.colorbar.kws.ax = self.currentFig.axes
                self.currentFig.colorbar = self.currentFig.figure.colorbar( **vars(self.colorbar.kws) )
                self.currentFig.colorbar.set_label( label = ", ".join(ccolnames) )

        ############################################################################################################################
        #### set axes scales
        ############################################################################################################################

        if self._xscale is not None: self.currentFig.axes.set_xscale(self._xscale)

        if self._yscale is not None: self.currentFig.axes.set_yscale(self._yscale)

        if self._zscale is not None and self._type.is3d: self.currentFig.axes.set_zscale(self._zscale)

        ############################################################################################################################
        #### set axes limits
        ############################################################################################################################

        if self._xlimit is not None:
            currentLim = list(self.currentFig.axes.get_xlim())
            if self._xlimit[0] is not None: currentLim[0] = self._xlimit[0]
            if self._xlimit[1] is not None: currentLim[1] = self._xlimit[1]
            self.currentFig.axes.set_xlim(currentLim)

        if self._ylimit is not None:
            currentLim = list(self.currentFig.axes.get_ylim())
            if self._ylimit[0] is not None: currentLim[0] = self._ylimit[0]
            if self._ylimit[1] is not None: currentLim[1] = self._ylimit[1]
            self.currentFig.axes.set_ylim(currentLim)

        if self._zlimit is not None and self._type.is3d:
            currentLim = list(self.currentFig.axes.get_zlim())
            if self._zlimit[0] is not None: currentLim[0] = self._zlimit[0]
            if self._zlimit[1] is not None: currentLim[1] = self._zlimit[1]
            self.currentFig.axes.set_zlim(currentLim)

        ############################################################################################################################
        #### add axes labels
        ############################################################################################################################

        if self._xlabel is None:
            if xcolindexlen>1:
                self.currentFig.axes.set_xlabel("Variable Values")
            else:
                self.currentFig.axes.set_xlabel(xcolnames[0])
        else:
            self.currentFig.axes.set_xlabel(self._xlabel)

        if self._ylabel is None:
            if ycolindexlen>1:
                self.currentFig.axes.set_ylabel("Variable Values")
            else:
                self.currentFig.axes.set_ylabel(ycolnames[0])
        else:
            self.currentFig.axes.set_ylabel(self._ylabel)

        if self._type.is3d:
            if self._zlabel is None:
                if zcolindexlen>1:
                    self.currentFig.axes.set_zlabel("Variable Values")
                else:
                    self.currentFig.axes.set_zlabel(zcolnames[0])
            else:
                self.currentFig.axes.set_zlabel(self._zlabel)

        ############################################################################################################################
        #### set legend and other BasePlot properties
        ############################################################################################################################

        self._finalizeBasePlot()

        if not self._type.is3d: self.target.currentFig.axes = self.currentFig.axes

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the LineScatterPlot class:" + newline \
                    + newline \
                    + self.__doc__ \
                    + newline \
                    + "Here is the help information on the parent BasePlot class:" + newline \
                    + newline \
                    + super().__doc__
        return docString

    ################################################################################################################################
    #### helpme
    ################################################################################################################################

    def helpme(self, topic=None):
        """

        Print the documentation for the input string topic. 
        If the topic does not exist, the documentation for
        the object will be printed.
     
            **Parameters**
         
                topic (optional)

                A string containing the name of the object 
                for which help is needed.

            **Returns**
         
                None
         
            **Example**

                .. code-block:: python
                    :linenos:

                    helpme()
                    helpme("make")
                    helpme("helpme")
                    helpme("getLogLinSpace")

        """

        try:
            exec("print(self."+topic+".__doc__)")
        except:
            print(self._getDocString())
        return None

    ################################################################################################################################
